from __future__ import annotations

import asyncio
import datetime
import os
import random
import time
from io import BytesIO
from typing import TYPE_CHECKING, Literal, Optional

import aiofiles
import aiohttp
import ujson

from .assets import Asset
from .message import Message
from .permissions import Permission
from .webhooks import Webhook

if TYPE_CHECKING:
    from ..api import HttpClient
    from ..bot import Bot
    from .users import User


class PermissionOverwrite:
    def __init__(self, payload: dict, bot: Bot):
        self.bot: Bot = bot
        self.http = bot.http
        self.update(payload)

    def update(self, payload: dict):
        self.id: str = payload["id"]
        self.type: int = payload["type"]
        self.allow: Permission = Permission(payload["allow"], self.bot)
        self.deny: Permission = Permission(payload["deny"], self.bot)


class SlashCommand:
    def __init__(self, payload: dict, bot: Bot) -> None:
        self.bot = bot
        self.http = bot.http
        self.update(payload)

    def update(self, payload):

        self.id = payload.get("id")
        self.type = payload.get("type")
        self.application_id = payload.get("application_id")
        self.guild_id = payload.get("guild_id")
        self.version = payload.get("version")
        self.name = payload.get("name")
        self.description = payload.get("description")
        self.description_default = payload.get("description_default")
        self.options = [
            SubCommandOption(option, self) for option in payload.get("options", [])
        ]
        self.my_options = []
        self.integration_types = payload.get("integration_types")
        self.raw_data = payload
        self.required = payload.get("required", False)

    def add_value(self, name: str, value):
        new = {}

        if type(value) == list:
            for option in self.options:
                if option.name == name:
                    new["name"] = name
                    new["type"] = option.type
                    new["options"] = value
            return self.my_options.append(new)

        for option in self.options:
            if option.name == name:
                new["name"] = name
                new["value"] = value
                new["type"] = option.type
        return self.my_options.append(new)

    def get_option(self, name: str) -> Optional[SubCommandOption]:
        for option in self.options:
            if option.name == name:
                return option


class SubCommandOption(SlashCommand):
    def __init__(self, payload: dict, main: SlashCommand) -> None:
        self.cmd = main
        self.update(payload)

    def update(self, payload):
        self.opt_options = payload.get("options", [])
        self.name = payload.get("name")
        self.description = payload.get("description")
        self.required = payload.get("required", False)
        self.type = payload.get("type")
        self.my_options = []

    def add_value(self, name: str, value):
        new = {}
        for option in self.opt_options:
            if option.get("name", "") == name:
                new["name"] = name
                new["value"] = value
                new["type"] = option["type"]
        self.my_options.append(new)

    def reconstruct(self):
        self.cmd.add_value(self.name, self.my_options)


class Channel:
    def __init__(self, payload: dict, bot: Bot):
        self.bot: Bot = bot
        self.http: HttpClient = bot.http
        self.update(payload)

    def update(self, payload):
        self.id = payload["id"]
        self.type = payload["type"]
        self.flags = payload.get("flags")
        self.last_message_id = payload.get("last_message_id")
        self.guild_id = payload.get("guild_id")

    @property
    def guild(self):
        return self.bot.fetch_guild(self.guild_id)

    async def delete(self):
        await self.http.request("delete", "/channels/" + self.id, json={})


class Callable(Channel):
    def __init__(self, payload: dict, bot: Bot):
        self.bot: Bot = bot
        self.http: HttpClient = bot.http
        super().__init__(payload, self.bot)
        super().update(payload)
        self.update(payload)

    def update(self, payload):
        pass

    async def call(self):
        if self.type in (1, 3):
            await self.bot.gateway.call(self.id, None)
        else:
            await self.bot.gateway.call(self.id, self.guild_id)
        if self.guild_id is None:
            await self.ring()

        await asyncio.sleep(1.5)
        if self.bot.gateway.voice:
            self.bot.gateway.voice.channel = self.id
            asyncio.create_task(self.bot.gateway.voice.start())

    async def video_call(self):
        if self.type in (1, 3):
            await self.bot.gateway.video_call(self.id, None)
        else:
            await self.bot.gateway.video_call(self.id, self.guild_id)
        if self.guild_id is None:
            await self.ring()

        await asyncio.sleep(1.5)
        if self.bot.gateway.voice:
            self.bot.gateway.voice.channel = self.id
            asyncio.create_task(self.bot.gateway.voice.start())

    async def ring(self):
        await self.http.request(
            "POST", f"/channels/{self.id}/call/ring", json={"recipients": None}
        )

    async def leave_call(self):
        await self.bot.gateway.leave_call()
        if self.bot.gateway.voice:
            self.bot.gateway.voice.channel = None
            await self.bot.gateway.voice.close()


class Messageable(Channel):
    def __init__(self, payload: dict, bot: Bot):
        self.bot: Bot = bot
        self.http: HttpClient = bot.http
        super().__init__(payload, self.bot)
        super().update(payload)
        self.update(payload)

    def __repr__(self):
        return f"<{self.__class__.__name__} id={self.id}>"

    def __str__(self):
        if hasattr(self, "name"):
            return self.name
        else:
            return self.recipient.username

    def update(self, payload):
        self.id: str = payload["id"]
        self.type: int = int(payload["type"])
        self.flags = payload.get("flags")
        self.last_message_id: Optional[str] = payload.get("last_message_id")

    def calc_nonce(self, date="now"):
        if date == "now":
            unixts = time.time()
        else:
            unixts = time.mktime(date.timetuple())
        return (int(unixts) * 1000 - 1420070400000) * 4194304

    @property
    def nonce(self) -> int:
        return self.calc_nonce()

    async def delayed_delete(self, message, time):
        await asyncio.sleep(time)
        await message.delete()

    async def get_slash_commands(
        self, name: Optional[str] = None
    ) -> Optional[SlashCommand] | Optional[list[SlashCommand]]:
        if self.guild_id is not None:
            json = await self.http.request(
                "GET", f"/guilds/{self.guild_id}/application-command-index"
            )
            if json is not None:
                cmds = []
                for cmd in json["application_commands"]:
                    if name is not None:
                        if cmd["name"] != name:
                            continue

                    if cmd.get("guild_id") is None:
                        cmd.update({"guild_id": self.guild_id})
                    cmds.append(SlashCommand(cmd, self.bot))

                if len(cmds) == 1:
                    return cmds[0]
                else:
                    return cmds

    async def trigger_slash(self, cmd: SlashCommand):
        if self.guild_id is not None:
            # print(cmd.my_options)

            json = {
                "type": 2,
                "application_id": cmd.application_id,
                "guild_id": self.guild_id,
                "channel_id": self.id,
                "session_id": self.bot.session_id,
                "data": {
                    "version": cmd.version,  # version not appending
                    "id": cmd.id,
                    "name": cmd.name,
                    "type": cmd.type,
                    "options": cmd.my_options,
                    "application_command": cmd.raw_data,
                },
            }
            json = ujson.dumps(json)
            boundary = f"----WebkitFormBoundaryAbCdEfGhIjKlmNoP"
            data = f'--{boundary}\r\nContent-Disposition: form-data; name="payload_json"\r\n\r\n{json}\r\n--{boundary}--'

            json = await self.http.request(
                "POST",
                "/interactions",
                headers={"content-type": f"multipart/form-data; boundary={boundary}"},
                data=data,
            )

    async def create_webhook(self, name: str):
        json = await self.http.request(
            "POST", f"/channels/{self.id}/webhooks", json={"name": name}
        )
        if json is not None:
            return Webhook(json, self.bot)

    async def typing(self):
        await self.http.request("POST", f"/channels/{self.id}/typing")

    async def upload_image(self, paths: list) -> list[dict[str, int | str]]:
        files = []
        id = 0
        for path in paths:
            if isinstance(path, (bytearray, bytes)):
                files.append(
                    {
                        "file_size": len(path),
                        "filename": f"{random.randint(1, 25555)}.png",
                        "id": id,
                    }
                )
            elif isinstance(path, BytesIO):
                files.append(
                    {
                        "file_size": path.getbuffer().nbytes,
                        "filename": f"{random.randint(1, 25555)}.png",
                        "id": id,
                    }
                )
            else:
                async with aiofiles.open(path, "rb") as f:
                    file = await f.read()
                    size = len(file)
                    name = os.path.basename(path)
                files.append({"file_size": size, "filename": f"{name}", "id": id})
            id += 1

        json = await self.http.request(
            "post", f"/channels/{self.id}/attachments", json={"files": files}
        )

        items = []
        for key, atch in enumerate(json["attachments"]):
            upload_url = atch["upload_url"]
            id = atch["id"]
            upload_filename = atch["upload_filename"]
            async with aiohttp.ClientSession() as session:
                if isinstance(paths[key], BytesIO):
                    file = paths[key].getvalue()
                elif isinstance(paths[key], (bytes, bytearray)):
                    file = paths[key]
                else:
                    async with aiofiles.open(paths[key], "rb") as f:
                        file = await f.read()
                async with session.put(upload_url, data=file):
                    pass
            items.append(
                {
                    "uploaded_filename": upload_filename,
                    "filename": os.path.basename(upload_filename),
                    "id": id,
                }
            )
        return items

    async def send(
        self,
        content: str = "",
        files: Optional[list[str]] = None,
        delete_after: Optional[int] = None,
        tts: bool = False,
    ) -> Optional[Message]:
        data = {"content": content, "tts": tts, "nonce": self.nonce}

        if files is not None and len(files) > 0:

            uploaded_files = await self.upload_image(files)
            data.update({"attachments": uploaded_files})

        if self.type in (1, 3):
            headers = {"referer": f"https://canary.discord.com/channels/@me/{self.id}"}
        else:
            headers = {
                "referer": f"https://canary.discord.com/channels/{self.guild_id}/{self.id}"
            }
        json = await self.http.request(
            "POST", f"/channels/{self.id}/messages", headers=headers, json=data
        )

        if json is not None:
            if delete_after != None:
                msg = Message(json, self.bot)
                await asyncio.create_task(self.delayed_delete(msg, delete_after))
                return msg

            return Message(json, self.bot)

    async def delete(self) -> Optional[Messageable]:
        if self.type in (1, 3):
            json = await self.http.request(
                "DELETE",
                f"/channels/{self.id}?silent=false",
                headers={
                    "referer": f"https://canary.discord.com/channels/@me/{self.id}"
                },
            )
        else:
            json = await self.http.request(
                "DELETE",
                f"/channels/{self.id}",
                headers={
                    "referer": f"https://canary.discord.com/channels/{self.guild_id}/{self.id}"
                },
            )

    async def history(self, limit: int = 50, bot_user_only: bool = False):
        n = 0
        if self.type in (1, 3):
            headers = {"referer": f"https://canary.discord.com/channels/@me/{self.id}"}
        else:
            headers = {
                "referer": f"https://canary.discord.com/channels/{self.guild_id}/{self.id}"
            }
        json = await self.http.request(
            "GET", f"/channels/{self.id}/messages?limit=50", headers=headers
        )
        msgs = []
        if json is None:
            return msgs

        for message in json:
            message.update({"guild_id": self.guild_id})
            msg = Message(message, self.bot)

            if bot_user_only:
                if msg.author.id == self.bot.user.id:
                    msgs.append(msg)
            else:
                msgs.append(msg)
            self.bot.cached_messages[msg.id] = msg
            n += 1

        if len(json) == 0:
            return msgs

        while n < limit:
            if json is None:
                return msgs

            if len(json) == 0:
                return msgs
            last = msgs[-1]
            json = await self.http.request(
                "GET",
                f"/channels/{self.id}/messages?before={last.id}&limit=50",
                headers=headers,
            )

            for message in json:
                message.update({"guild_id": self.guild_id})
                msg = Message(message, self.bot)
                if bot_user_only:
                    if msg.author.id == self.bot.user.id:
                        msgs.append(msg)
                else:
                    msgs.append(msg)
                self.bot.cached_messages[msg.id] = msg
                n += 1

        return msgs[:limit]

    async def search_base(
        self,
        content: Optional[str] = None,
        has: Optional[
            Literal["link", "embed", "file", "video", "image", "sound"]
        ] = None,
        from_: Optional[str] = None,
        mentions: Optional[str] = None,
        max: Optional[int] = None,
        min: Optional[int] = None,
        offset: Optional[int] = None,
    ) -> Optional[list[Message]]:
        params = {
            "content": content,
            "has": has,
            "from": from_,
            "mentions": mentions,
            "max": max,
            "min": min,
            "offset": offset,
        }
        params = {k: v for k, v in params.items() if v is not None}
        json = await self.http.request(
            "GET", f"/guilds/{self.id}/messages/search", params=params
        )

        if json is not None:
            for messages in json["messages"]:
                for message in messages:
                    message.update({"guild_id": self.guild_id})
                    msg = Message(message, self.bot)

                    self.bot.cached_messages[message.id] = msg
            return [
                Message(message, self.bot)
                for messages in json["messages"]
                for message in messages
            ]

    async def search(
        self,
        content: Optional[str] = None,
        has: Optional[
            Literal["link", "embed", "file", "video", "image", "sound"]
        ] = None,
        from_: Optional[str] = None,
        mentions: Optional[str] = None,
        max: Optional[int] = None,
        min: Optional[int] = None,
        amount: int = 25,
    ) -> list[Message]:
        n = 0
        msgs = []
        while n < amount:
            msg = await self.search_base(
                content=content,
                has=has,
                from_=from_,
                mentions=mentions,
                max=max,
                min=min,
                offset=n,
            )
            if msg is not None:
                if len(msg) == 0:
                    break
                n += len(msg)
                msgs.extend(msg)
        return msgs

    async def get_webhooks(self) -> Optional[list[Webhook]]:
        json = await self.http.request("GET", f"/channels/{self.id}/webhooks")
        if json is not None:
            return [Webhook(webhook, self.bot) for webhook in json]

    async def purge(self, amount: int):
        msgs = await self.history(amount, bot_user_only=True)
        for i in range(0, len(msgs), 2):
            await asyncio.gather(*(msg.delete() for msg in msgs[i : i + 2]))


class DMChannel(Messageable, Callable):
    def __init__(self, payload: dict, bot: Bot):

        super().__init__(payload, bot)

        super().update(payload)

        self.bot = bot
        self.http = bot.http

        self.update(payload)

    @property
    def recipient(self):
        return self.bot.fetch_user(self.recipient_id)

    def update(self, payload: dict):

        for recipien in payload.get("recipients", payload.get("recipient_ids", [])):
            if isinstance(recipien, str):
                self.recipient_id = recipien
            else:
                self.recipient_id = recipien["id"]

        self.is_spam: Optional[bool] = payload.get("is_spam")

    def get_user(self):
        self.recipient = self.bot.fetch_user(self.recipient_id)


class GroupChannel(Messageable, Callable):
    def __init__(self, payload: dict, bot: Bot):
        super().update(payload)
        super().__init__(payload, bot)
        self.bot = bot
        self.http = bot.http
        self.update(payload)

    def update(self, payload: dict):
        self.recipient: list[Optional[User]] = (
            [self.bot.fetch_user(user) for user in payload["recipient_ids"]]
            if payload.get("recipient_ids") is not None
            else []
        )
        self.is_spam: Optional[bool] = payload.get("is_spam")
        self.icon: Optional[Asset] = (
            Asset(self.id, payload["icon"]).from_icon()
            if payload.get("icon") is not None
            else None
        )
        self.name: Optional[str] = payload.get("name")
        self.last_pin_timestamp: Optional[int] = payload.get("last_pin_timestamp")

    async def leave(self):
        await self.http.request("DELETE", "/channels/" + self.id)

    async def edit(self, name: str, icon_url: Optional[str] = None):
        payload = {"name": name}
        if icon_url is not None:
            payload["icon"] = await self.http.encode_image(icon_url)

        await self.http.request("PATCH", "/channels/" + self.id, json=payload)


class TextChannel(Messageable):
    """
    This class is used to represent a text channel in Discord.
    """

    def __init__(self, payload: dict, bot: Bot):
        self.bot = bot
        self.http = bot.http
        self.update(payload)
        super().update(payload)
        super().__init__(payload, bot)

    def update(self, payload):
        self.nsfw = payload.get("nsfw")
        self.guild_id = payload.get("guild_id")
        self.category_id = payload.get("parent_id")
        self.position = payload.get("position")
        self.rate_limit_per_user = payload.get("rate_limit_per_user")
        self.name = payload.get("name")
        self.last_pin_timestamp = payload.get("last_pin_timestamp")

        self.permission_overwrites = [PermissionOverwrite(overwrite, self.bot) for overwrite in payload.get("permission_overwrites", [])]  # type: ignore

    async def edit(
        self, name: str = "", topic: str = "", nsfw: bool = False, timeout: int = 0
    ):
        self.name = name
        self.nsfw = nsfw
        self.rate_limit_per_user = timeout

        await self.http.request(
            "patch",
            "/channels/" + self.id,
            json={
                "name": name,
                "type": 0,
                "topic": topic,
                "bitrate": 64000,
                "user_limit": 0,
                "nsfw": nsfw,
                "flags": 0,
                "rate_limit_per_user": timeout,
            },
        )


class VoiceChannel(Messageable, Callable):
    def __init__(self, payload: dict, bot: Bot):
        self.bot = bot
        self.http = bot.http
        self.update(payload)
        super().update(payload)
        super().__init__(payload, bot)

    def update(self, payload):
        self.guild_id = payload.get("guild_id")
        self.category_id = payload.get("parent_id")
        self.position = payload.get("position")
        self.permission_overwrites = [PermissionOverwrite(overwrite, self.bot) for overwrite in payload.get("permission_overwrites", [])]  # type: ignore
        self.user_limit = payload.get("user_limit")
        self.topic = payload.get("topic")
        self.rtc_region = payload.get("rtc_region")
        self.slowdown = payload.get("rate_limit_per_user")
        self.nsfw = payload.get("nsfw")
        self.name = payload.get("name")
        self.icon_emoji = payload.get("icon_emoji")
        self.bitrate = payload.get("bitrate")


class Category(Messageable):
    def __init__(self, payload: dict, bot: Bot):
        self.bot = bot
        self.http = bot.http
        self.update(payload)
        super().update(payload)
        super().__init__(payload, bot)

    def update(self, payload):
        self.name = payload.get("name")
        self.guild_id = payload.get("guild_id")
        self.position = payload.get("position")
        self.permission_overwrites = [PermissionOverwrite(overwrite, self.bot) for overwrite in payload.get("permission_overwrites", [])]  # type: ignore


class Announcement(Messageable):
    def __init__(self, payload: dict, bot: Bot):
        self.bot = bot
        self.http = bot.http
        self.update(payload)
        super().update(payload)
        super().__init__(payload, bot)

    def update(self, payload):
        self.name = payload.get("name")
        self.guild_id = payload.get("guild_id")
        self.position = payload.get("position")
        self.permission_overwrites = [PermissionOverwrite(overwrite, self.bot) for overwrite in payload.get("permission_overwrites", [])]  # type: ignore


class AnnouncementThread(Messageable):
    def __init__(self, payload: dict, bot: Bot):
        self.bot = bot
        self.http = bot.http
        self.update(payload)
        super().update(payload)
        super().__init__(payload, bot)

    def update(self, payload):
        self.name = payload.get("name")
        self.guild_id = payload.get("guild_id")
        self.position = payload.get("position")
        self.permission_overwrites = [PermissionOverwrite(overwrite, self.bot) for overwrite in payload.get("permission_overwrites", [])]  # type: ignore


class PublicThread(Messageable):
    def __init__(self, payload: dict, bot: Bot):
        self.bot = bot
        self.http = bot.http
        self.update(payload)
        super().update(payload)
        super().__init__(payload, bot)

    def update(self, payload):
        self.name = payload.get("name")
        self.guild_id = payload.get("guild_id")
        self.position = payload.get("position")
        self.permission_overwrites = [PermissionOverwrite(overwrite, self.bot) for overwrite in payload.get("permission_overwrites", [])]  # type: ignore


class PrivateThread(Messageable):
    def __init__(self, payload: dict, bot: Bot):
        self.bot = bot
        self.http = bot.http
        self.update(payload)
        super().update(payload)
        super().__init__(payload, bot)

    def update(self, payload):
        self.name = payload.get("name")
        self.guild_id = payload.get("guild_id")
        self.position = payload.get("position")
        self.permission_overwrites = [PermissionOverwrite(overwrite, self.bot) for overwrite in payload.get("permission_overwrites", [])]  # type: ignore


class StageChannel(Messageable):
    def __init__(self, payload: dict, bot: Bot):
        self.bot = bot
        self.http = bot.http
        self.update(payload)
        super().update(payload)
        super().__init__(payload, bot)

    def update(self, payload):
        self.last_message_id: Optional[str] = payload.get("last_message_id")
        self.name = payload.get("name")
        self.guild_id = payload.get("guild_id")
        self.position = payload.get("position")
        self.permission_overwrites = [PermissionOverwrite(overwrite, self.bot) for overwrite in payload.get("permission_overwrites", [])]  # type: ignore


class Directory(Messageable):
    def __init__(self, payload: dict, bot: Bot):
        self.bot = bot
        self.http = bot.http
        self.update(payload)
        super().update(payload)
        super().__init__(payload, bot)

    def update(self, payload):
        self.name = payload.get("name")
        self.guild_id = payload.get("guild_id")
        self.position = payload.get("position")
        self.permission_overwrites = [PermissionOverwrite(overwrite, self.bot) for overwrite in payload.get("permission_overwrites", [])]  # type: ignore


class ForumChannel(Messageable):
    def __init__(self, payload: dict, bot: Bot):
        self.bot = bot
        self.http = bot.http
        self.update(payload)
        super().update(payload)
        super().__init__(payload, bot)

    def update(self, payload):
        self.name = payload.get("name")
        self.guild_id = payload.get("guild_id")
        self.position = payload.get("position")
        self.topic = payload.get("topic")
        self.template = payload.get("template")
        self.slowdown = payload.get("rate_limit_per_user")
        self.category_id = payload.get("category_id")
        self.nsfw = payload.get("nsfw")
        self.default_thread_rate_limit_per_user = payload.get(
            "default_thread_rate_limit_per_user"
        )
        self.default_sort_order = payload.get("default_sort_order")
        self.default_reaction_emoji = payload.get("default_reaction_emoji")
        self.default_forum_layout = payload.get("default_forum_layout")
        self.available_tags = payload.get("available_tags")

        self.permission_overwrites = [PermissionOverwrite(overwrite, self.bot) for overwrite in payload.get("permission_overwrites", [])]  # type: ignore


class MediaChannel(Messageable):
    def __init__(self, payload: dict, bot: Bot):
        self.bot = bot
        self.http = bot.http
        self.update(payload)
        super().update(payload)
        super().__init__(payload, bot)

    def update(self, payload):
        self.name = payload.get("name")
        self.guild_id = payload.get("guild_id")
        self.position = payload.get("position")
        self.permission_overwrites = [PermissionOverwrite(overwrite, self.bot) for overwrite in payload.get("permission_overwrites", [])]  # type: ignore


class Convert(Messageable):
    def __new__(cls, payload: dict, bot: Bot) -> Messageable:

        tpe = payload["type"]

        if tpe == 0:
            return TextChannel(payload, bot)
        if tpe == 1:
            return DMChannel(payload, bot)
        if tpe == 2:
            return VoiceChannel(payload, bot)
        if tpe == 3:
            return GroupChannel(payload, bot)
        if tpe == 4:
            return Category(payload, bot)
        if tpe == 5:
            return Announcement(payload, bot)
        if tpe == 10:
            return AnnouncementThread(payload, bot)
        if tpe == 11:
            return PublicThread(payload, bot)
        if tpe == 12:
            return PrivateThread(payload, bot)
        if tpe == 13:
            return StageChannel(payload, bot)
        if tpe == 14:
            return Directory(payload, bot)
        if tpe == 15:
            return ForumChannel(payload, bot)
        if tpe == 16:
            return MediaChannel(payload, bot)
        return TextChannel(payload, bot)



"""Where command handling and logging reside. This also is where I wrote extensions in commands.py"""
from .command import (Command, CommandCollection, Context, Event, Extender,
                      Extension, ExtensionCollection)
from .logging import logging


"""Where Selfcord interacts with discords API directly, using discord gateway (websockets) and http requests. This is also where events are located."""
from .gateway import Gateway
from .http import HttpClient
from .voice import *
from .errors import *
from .ipc import DiscordIPC
from .activity import Activity


from websockets.client import connect
import ujson
from aioconsole import aprint
from typing import Optional
import time
import asyncio
import socket
import io
import struct
import ctypes
import array

try:
    import opuslib
    import nacl.secret
    import nacl.utils
except ImportError:
    pass



SAMPLING_RATE = 48000
CHANNELS = 2
FRAME_LENGTH = 20  # in milliseconds
SAMPLE_SIZE = struct.calcsize("h") * CHANNELS
SAMPLES_PER_FRAME = int(SAMPLING_RATE / 1000 * FRAME_LENGTH)
FRAME_SIZE = SAMPLES_PER_FRAME * SAMPLE_SIZE



class Source(io.BufferedIOBase):
    def __init__(self, source: bytes) -> None:
        self.source: io.BufferedIOBase = io.BytesIO(source)

    def read(self):
        data = self.source.read(FRAME_SIZE)
        if len(data) != FRAME_SIZE:
            return b''
        return data
    
    def __len__(self):
        return self.source.getbuffer().nbytes - self.source.tell()
    


    







class Voice:
    IDENTIFY = 0
    UDP_SELECT = 1
    READY = 2
    HEARTBEAT = 3
    SESSION_DESCRIPTION = 4
    HEARTBEAT_ACK = 6
    HELLO = 8


    def __init__(
        self,
        bot,
        voice_token: str,
        endpoint: str,
        server_id: str
    ) -> None:
        self.bot = bot
        self.voice_token = voice_token
        self.session_id = bot.session_id
        self.endpoint = endpoint
        self.server_id = server_id
        self.alive = False
        self.channel = None
        self.ws = None
        self.mode = "xsalsa20_poly1305_lite"
        self._lite_nonce = 0
        self.sequence = 0
        self.timestamp = 0
        try:
            self.encoder = opuslib.Encoder(SAMPLING_RATE, CHANNELS, application=2049) 
        except:
            pass
        self.socket: Optional[socket.socket] = None


    async def send(self, payload: dict):
        if self.ws:
            await self.ws.send(ujson.dumps(payload))

    async def recv(self):
        if self.ws:
            item = await self.ws.recv()
            json = ujson.loads(item)
      
            op = json['op']
            data = json['d']

            if op == self.HELLO:
                interval = data['heartbeat_interval'] / 1000.0
                await self.identify()
                asyncio.create_task(
                    self.heartbeat(interval)
                )
            
            if op == self.HEARTBEAT_ACK:
                pass


            if op == self.SESSION_DESCRIPTION:
                await self.handle_session_description(data)


            if op == self.READY:
                await self.handle_ready(data)

    async def handle_session_description(self, data: dict):
        self.secret_key = data['secret_key']
        self.audio_codec = data['audio_codec']
        self.media_session_id = data['media_session_id']

    

    async def handle_ready(self, data: dict):
        self.ssrc = data['ssrc']
        self.ip = data['ip']
        self.port = data['port']
        self.modes = data['modes']
        heartbeat_interval = data['heartbeat_interval']
        await self.ip_discovery()
        await self.udp_select()

    async def connect(self):
        self.ws = await connect(f"wss://{self.endpoint}", origin="https://discord.com")
        self.alive = True
    
    async def start(self):
        await self.connect()
        while self.alive:
            await self.recv()


        
    def encrypt_xsalsa20_poly1305_lite(self, header: bytes, data) -> bytes:
        box = nacl.secret.SecretBox(bytes(self.secret_key))
        nonce = bytearray(24)

        nonce[:4] = struct.pack('>I', self._lite_nonce)
        self.checked_add('_lite_nonce', 1, 4294967295)

        return header + box.encrypt(bytes(data), bytes(nonce)).ciphertext + nonce[:4]
    

    
    def encode(self, pcm_data: bytes):
        c_int16_ptr = ctypes.POINTER(ctypes.c_int16)
        max_len = len(pcm_data)
        pcm_ptr = ctypes.cast(pcm_data, c_int16_ptr) # type: ignore
        data = (ctypes.c_char * max_len)() # type: ignore

        ret = self.encoder.encode(pcm_data, FRAME_SIZE)
        return array.array('b', data[:len(ret)]).tobytes()

    
    async def send_audio_data(self, source: bytes):
        self.checked_add("sequence", 1, 65535)
        self.source = Source(source)
        start = time.perf_counter()
        loops = 0
        while len(self.source) != 0:
            
            data = self.source.read()
            loops += 1

            encoded = self.encode(data)
            packet = self.get_voice_packet(encoded)

            await self.speaking(speak=True)
            if self.socket:
                # await aprint("sending", len(data), f"LEFT: {len(self.source)}")
                self.socket.sendto(packet, (self.ip, self.port))

            next_time = start + (FRAME_LENGTH / 1000) * loops
            delay = max(0, (FRAME_LENGTH / 1000) +  (next_time - time.perf_counter()))
            await aprint(delay)
            await asyncio.sleep(delay)
            self.checked_add('timestamp', SAMPLES_PER_FRAME, 4294967295)
           
    
    def get_voice_packet(self, data):
        header = bytearray(12)

        # Formulate rtp header
        header[0] = 0x80
        header[1] = 0x78
        struct.pack_into('>H', header, 2, self.sequence)
        struct.pack_into('>I', header, 4, self.timestamp)
        struct.pack_into('>I', header, 8, self.ssrc)


        return self.encrypt_xsalsa20_poly1305_lite(header, data)


    
    def checked_add(self, attr: str, value: int, limit: int) -> None:
        val = getattr(self, attr)
        if val + value > limit:
            setattr(self, attr, 0)
        else:
            setattr(self, attr, val + value)


    async def speaking(self, speak: bool = False, priority = False):
        val = 0
        if speak:
            val += 1
        if priority:
            val += 4
        payload = {
            "op": 5,
            "d": {
               "speaking": int(val),
               "delay": 0,
               "ssrc": self.ssrc 
            }
        }
    
        await self.send(payload)

    async def identify(self):
        payload = {
            "op": self.IDENTIFY,
            "d": {
                "server_id": self.server_id,
                "user_id": self.bot.user.id,
                "session_id": self.session_id,
                "token": self.voice_token
            }
        }
        await self.send(payload)


    async def heartbeat(self, interval: int):
        while self.alive:
            await asyncio.sleep(interval)
            payload = {
                "op": self.HEARTBEAT, "d": int(time.time())
            }
            await self.send(payload)

    async def close(self):
        if self.ws:
            self.alive = False
            await self.ws.close()
            
    async def udp_select(self):
      
        payload = {
            "op": 1,
            "d": {
                "protocol": "udp",

                "data": {"address": self.my_ip, "port": self.my_port, "mode": self.mode},
            },
        }
        await self.send(payload)


    async def ip_discovery(self):
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)

        self.socket.setblocking(False)

        packet = bytearray(74)
        struct.pack_into(">H", packet, 0, 1)
        struct.pack_into(">H", packet, 2, 70)
        struct.pack_into(">I", packet, 4, self.ssrc)
        self.socket.sendto(packet, (self.ip, self.port))

        while True:
            try:
                data = self.socket.recv(74)
                break
            except:
                continue
        ip_start = 8
        ip_end = data.index(0, ip_start)
        self.my_ip = data[ip_start:ip_end].decode("ascii")
        self.my_port = struct.unpack_from(">H", data, len(data) - 2)[0]
        print(self.my_ip, self.my_port)
    
    


from __future__ import annotations
from typing import Self, Type


class Flags:
    def __init__(self, value: int):
        self.value: int = value

    @classmethod
    def from_value(cls, value: int):
        self = cls(value)
        self.value = value
        return self

    def __repr__(self) -> str:
        return f"{self.value}"


class Capabilities(Flags):
    
    @classmethod
    def default(cls):
        """Utilise default values for optimal anti-flag"""
        # Discum uses 4093
        # Desktop client uses 16381
        # Discord.py-self uses 8189
        return cls.from_value(8189)

    @classmethod
    def lazy_user_notes(cls: Type[Self]) -> Self:
        """Disable preloading of user notes in READY."""
        return cls.from_value(1 << 0)

    @classmethod
    def no_affine_user_ids(cls: Type[Self]) -> Self:
        """Disable implicit relationship updates."""
        return cls.from_value(1 << 1)

    @classmethod
    def versioned_read_states(cls: Type[Self]) -> Self:
        """Enable versioned read states (change READY ``read_state`` to an object with ``version``/``partial``)."""
        return cls.from_value(1 << 2)

    @classmethod
    def versioned_user_guild_settings(cls: Type[Self]) -> Self:
        """Enable versioned user guild settings (change READY ``user_guild_settings`` to an object with ``version``/``partial``)."""
        return cls.from_value(1 << 3)

    @classmethod
    def dedupe_user_objects(cls: Type[Self]) -> Self:
        """Enable dehydration of the READY payload (move all user objects to a ``users`` array and replace them in various places in the READY payload with ``user_id`` or ``recipient_id``, move member object(s) from initial guild objects to ``merged_members``)."""
        return cls.from_value(1 << 4)

    @classmethod
    def prioritized_ready_payload(cls: Type[Self]) -> Self:
        """Enable prioritized READY payload (enable READY_SUPPLEMENTAL, move ``voice_states`` and ``embedded_activities`` from initial guild objects and ``merged_presences`` from READY, as well as split ``merged_members`` and (sometimes) ``private_channels``/``lazy_private_channels`` between the events)."""
        # Requires self.dedupe_user_objects
        return cls.from_value(1 << 5 | 1 << 4)

    @classmethod
    def multiple_guild_experiment_populations(cls: Type[Self]) -> Self:
        """Handle multiple guild experiment populations (change the fourth entry of arrays in the ``guild_experiments`` array in READY to have an array of population arrays)."""
        return cls.from_value(1 << 6)

    @classmethod
    def non_channel_read_states(cls: Type[Self]) -> Self:
        """Handle non-channel read states (change READY ``read_state`` to include read states tied to server events, server home, and the mobile notification center)."""
        return cls.from_value(1 << 7)

    @classmethod
    def auth_token_refresh(cls: Type[Self]) -> Self:
        """Enable auth token refresh (add ``auth_token?`` to READY; this is sent when Discord wants to change the client's token, and was used for the mfa. token migration)."""
        return cls.from_value(1 << 8)

    @classmethod
    def user_settings_proto(cls: Type[Self]) -> Self:
        """Disable legacy user settings (remove ``user_settings`` from READY and stop sending USER_SETTINGS_UPDATE)."""
        return cls.from_value(1 << 9)

    @classmethod
    def client_state_v2(cls: Type[Self]) -> Self:
        """Enable client caching v2 (move guild properties in guild objects to a ``properties`` subkey and add ``data_mode`` and ``version`` to the objects, as well as change ``client_state`` in IDENTIFY)."""
        return cls.from_value(1 << 10)

    @classmethod
    def passive_guild_update(cls: Type[Self]) -> Self:
        """Enable passive guild update (replace ``CHANNEL_UNREADS_UPDATE`` with ``PASSIVE_UPDATE_V1``, a similar event that includes a ``voice_states`` array and a ``members`` array that includes the members of aforementioned voice states)."""
        return cls.from_value(1 << 11)



from __future__ import annotations
from typing import Optional, TYPE_CHECKING
from .assets import Asset
from .message import Message

if TYPE_CHECKING:
    from ..bot import Bot
    from .channels import Channel
    from .guild import Guild


# TODO: Implement webhooks


class Webhook:
    def __init__(self, payload: dict, bot: Bot):
        self.bot = bot
        self.http = bot.http
        self.update(payload)

    @property
    def url(self) -> str:
        return f"https://discord.com/api/webhooks/{self.id}/{self.token}"
    
    @property
    def avatar(self) -> Asset:
        return Asset(self.id, self.avatar_hash)
    
    @property
    def channel(self) -> Optional[Channel]:
        return self.bot.get_channel(self.channel_id)
    
    @property
    def guild(self) -> Optional[Guild]:
        return self.bot.fetch_guild(self.guild_id)
    


    def update(self, payload: dict):
        self.id = payload.get("id")
        self.type = payload.get("type")
        self.guild_id = payload.get("guild_id")
        self.channel_id = payload.get("channel_id")
        self.user = payload.get("user")
        self.name = payload.get("name")
        self.avatar_hash = payload.get("avatar", "")

        self.token = payload.get("token")
        self.application_id = payload.get("application_id")
        self.source_guild = payload.get("source_guild")
        self.source_channel = payload.get("source_channel")

    async def delete(self):
        await self.http.request(
            "DELETE",
            f"/webhooks/{self.id}/{self.token}"
        )

    async def send(self, content: Optional[str] = None):
        json = await self.http.request(
            "POST",
            f"/webhooks/{self.id}/{self.token}",
            json={
                "content": content,
            
            }
        )
        if json is not None:
            return Message(json, self.bot)
    



from .users import User, Client, Member
from .flags import Flags, Capabilities
from .guild import Guild, Role, Sticker, Emoji
from .assets import Asset
from .channels import (
    TextChannel,
    DMChannel,
    VoiceChannel,
    GroupChannel,
    Category,
    Announcement,
    AnnouncementThread,
    PublicThread,
    PrivateThread,
    Directory,
    StageChannel,
    ForumChannel,
    MediaChannel,
    Convert,
    Messageable,
    Callable
)
from .webhooks import Webhook
from .message import Message, MessageAck
from .event_models import PresenceUpdate, MessageAddReaction, MemberListUpdate, CallCreate


import asyncio
import sys
import os
import tempfile
import json
import struct
from typing import Optional
import time


# Thanks to these people
# Pretty much took most of their code
# https://github.com/qwertyquerty/pypresence

class DiscordIPC:
    def __init__(self, application_id: str, pid: int = os.getpid()) -> None:
        self.application_id = application_id
        self.pid = pid

    @property
    def path(self):
        ipc = 'discord-ipc-'
        

        if sys.platform in ('linux', 'darwin'):
            tempdir = (os.environ.get('XDG_RUNTIME_DIR') or tempfile.gettempdir())
            paths = ['.', 'snap.discord', 'app/com.discordapp.Discord', 'app/com.discordapp.DiscordCanary']
        elif sys.platform == 'win32':
            tempdir = r'\\?\pipe'
            paths = ['.']
        else:
            return
        
        for path in paths:
            full_path = os.path.abspath(os.path.join(tempdir, path))
            if sys.platform == 'win32' or os.path.isdir(full_path):
                for entry in os.scandir(full_path):
                    if entry.name.startswith(ipc) and os.path.exists(entry):
                        return entry.path
                    
    async def read_output(self):
        try:
            msg = await asyncio.wait_for(self.read.read(8), 20)
            status_code, length = struct.unpack('<II', msg[:8])
            data = await asyncio.wait_for(self.read.read(length), 20)
        except (BrokenPipeError, struct.error):
            raise BrokenPipeError
        except asyncio.TimeoutError:
            return
        payload = json.loads(data.decode('utf-8'))
        if payload["evt"] == "ERROR":
            raise RuntimeError(payload["data"]["message"])
        return payload


    def send(self, op: int, payload: dict):
        payload = json.dumps(payload)

        
        self.write.write(
            struct.pack(
                '<II',
                op,
                len(payload)
            ) + payload.encode('utf-8')
        )


    async def connect(self):
        self.read, self.write = await asyncio.open_unix_connection(self.path)

        self.send(0, {'v': 1, 'client_id': self.application_id})
        msg = await self.read.read(8)
        code, length = struct.unpack('<ii', msg)
        msg = json.loads((await self.read.read(length)))
        if "code" in msg:
            if msg['message'] == "Invalid Client Id":
                raise RuntimeError("You failed, wrong ID")
            
        asyncio.create_task(self.loop())

    def close(self):
        self.send(2, {'v': 1, 'client_id': self.application_id})
        self.write.close()
        self.loop.close()
            
    async def loop(self):
        while True:
            await self.read_output()

    
    async def update_presence(
        self,
        state: Optional[str] = None,
        details: Optional[str] = None,
        start: Optional[int] = None,
        end: Optional[int] = None,
        large_image: Optional[str] = None,
        large_text: Optional[str] = None,
        small_image: Optional[str] = None,
        small_text: Optional[str] = None,
        party_id: Optional[str] = None,
        party_size: Optional[list] = None,
        join: Optional[str] = None,
        match: Optional[str] = None,
        spectate: Optional[str] = None,
        buttons: Optional[list] = None,
        instance: bool = True, payload_override: Optional[dict] = None
    ):
        if payload_override:
            payload = payload_override
        else:
            if start:
                start = int(start)
            if end:
                end = int(end)

            activity = {
                "state": state,
                "details": details,
                "timestampts": {
                    "start": start,
                    "end": end
                },
                "assets": {
                    "large_image": large_image,
                    "large_text": large_text,
                    "small_image": small_image,
                    "small_text": small_text,
                },
                "party": {
                    "party_id": party_id,
                    "size": party_size
                },
                "secrets": {
                    "join": join,
                    "spectate": spectate,
                    "match": match
                },
                "buttons": buttons,
                "instance": instance
            }
            payload = {
                "cmd": "SET_ACTIVITY",
                "args": {
                    "pid": self.pid,
                    "activity": activity
                },
                "nonce": f"{time.time():.20f}",
            }
        self.send(1, payload)




from __future__ import annotations
import itertools

from typing import Optional
from .users import User, Status, Member
from .guild import Emoji

class PresenceUpdate:
    def __init__(self, payload: dict, bot) -> None:
        self.bot = bot
        self.http = bot.http
        self.update(payload)

    def update(self, payload: dict):
        self.user = payload.get("user")
        if self.user is not None:
            if self.user.get("username") is not None:
                self.user = User(self.user, self.bot)
            else:
                self.user = self.bot.fetch_user(self.user['id']) if self.bot.fetch_user(self.user['id']) is not None else self.user['id']
        self.status = payload.get("status")
        self.client_status = Status(payload['client_status']) if payload.get("client_status") is not None else payload.get("client_status")
        self.activities = payload.get("activities")
        self.broadcast = payload.get("broadcast")

                

class MessageAddReaction:
    def __init__(self, payload: dict, bot) -> None:
        self.bot = bot
        self.http = bot.http
        self.update(payload)

    def update(self, payload):
        self.burst = payload.get("burst", False)
        self.message_id = payload.get("message_id")
        self.channel_id = payload.get("channel_id")
        self.channel = self.bot.get_channel(self.channel_id)
        self.message = self.bot.fetch_message(self.message_id)
        self.message_author_id = payload.get("message_author_id")
        self.message_author = self.bot.fetch_user(self.message_author_id)
        self.user_id = payload.get("user_id")
        self.user = self.bot.fetch_user(self.user_id)
        self.type = payload.get("type")
        self.emoji = Emoji(payload['emoji'], self.bot) if payload.get("emoji") is not None else None

class MemberListUpdate:
    def __init__(self, payload: dict, bot) -> None:
        self.bot = bot
        self.http = bot.http
        self.update(payload)

    def update(self, payload):
        self.member_count = payload.get("member_count")
        self.id = payload.get("id")
        self.guild_id = payload.get("guild_id")

        self.groups = []
        self.ops = []
        for group, op in itertools.zip_longest(payload.get("groups", []), payload.get("ops", [])):
            if group is not None:
                group.update({"guild_id": self.guild_id})
                self.groups.append(group)

            if op is not None:
                op.update({"guild_id": self.guild_id})
                self.ops.append(op)

class MemberOp():
    def __init__(self, payload: dict, bot) -> None:
        self.bot = bot
        self.http = bot.http
        self.update(payload)

    def update(self, payload):
        self.op = payload.get("op")
        self.range = payload.get("range")
        self.items = payload.get("items", [])
        self.index = payload.get("index")
        self.guild_id = payload.get("guild_id")
        for item in self.items:
            member = item.get("member")
            if member is not None:
                guild = self.bot.fetch_guild(self.guild_id)
                member = Member(member, self.bot)
                guild.members.append(member)


class Group:
    def __init__(self, payload: dict, bot) -> None:
        self.bot = bot
        self.http = bot.http
        self.update(payload)

    def update(self, payload):
        self.group_id = payload.get("id")
        self.count = payload.get("count")

class CallCreate:
    def __init__(self, payload: dict, bot) -> None:
        self.bot = bot
        self.http = bot.http
        self.update(payload)

    @property
    def channel(self):
        return self.bot.get_channel(self.channel_id)
    
    def update(self, payload):
        self.voice_states = []
        for user in payload.get("voice_states", []):
            
            check_user = self.bot.fetch_user(user['user_id'])
            if check_user is None:
                self.voice_states.append(User(user, self.bot))
            else:
                check_user.partial_update(user)
                self.voice_states.append(check_user)
        self.ringing = payload.get("ringing", [])
        self.region = payload.get("region")
        self.channel_id = payload.get("channel_id")




from __future__ import annotations
from typing import Self


class Asset:
    def __init__(self, id: str, hash: str) -> None:
        self.url: str = ""
        self.update(id, hash)

    def update(self, id: str, hash: str):
        self.id: str = id
        self.hash: str = hash

    def __str__(self) -> str:
        return self.url

    def from_avatar(self):
        self.url = (
            f"https://cdn.discordapp.com/avatars/{self.id}/{self.hash}.png?size=4096"
        )
        if self.hash.startswith("a_"):
            self.url = f"https://cdn.discordapp.com/avatars/{self.id}/{self.hash}.gif?size=4096"
        return self

    def from_icon(self) -> Self:
        self.url = (
            f"https://cdn.discordapp.com/icons/{self.id}/{self.hash}.png?size=4096"
        )
        if self.hash.startswith("a_"):
            self.url = (
                f"https://cdn.discordapp.com/icons/{self.id}/{self.hash}.gif?size=4096"
            )
        return self

    @property
    def is_animated(self):
        if self.hash.startswith("a_"):
            return True
        return False



from __future__ import annotations

from typing import Optional

class Permission:
    CREATE_INSTANT_INVITE = 1 << 0
    KICK_MEMBERS = 1 << 1
    BAN_MEMBERS = 1 << 2
    ADMINISTRATOR = 1 << 3
    MANAGE_CHANNELS = 1 << 4
    MANAGE_GUILD = 1 << 5
    ADD_REACTIONS = 1 << 6
    VIEW_AUDIT_LOG = 1 << 7
    PRIORITY_SPEAKER = 1 << 8
    STREAM = 1 << 9
    VIEW_CHANNEL = 1 << 10
    SEND_MESSAGES = 1 << 11
    SEND_TTS_MESSAGES = 1 << 12
    MANAGE_MESSAGES = 1 << 13
    EMBED_LINKS = 1 << 14
    ATTACH_FILES = 1 << 15
    READ_MESSAGE_HISTORY = 1 << 16
    MENTION_EVERYONE = 1 << 17
    USE_EXTERNAL_EMOJIS = 1 << 18
    VIEW_GUILD_INSIGHTS = 1 << 19
    CONNECT = 1 << 20
    SPEAK = 1 << 21
    MUTE_MEMBERS = 1 << 22
    DEAFEN_MEMBERS = 1 << 23
    MOVE_MEMBERS = 1 << 24
    USE_VAD = 1 << 25
    CHANGE_NICKNAME = 1 << 26
    MANAGE_NICKNAMES = 1 << 27
    MANAGE_ROLES = 1 << 28
    MANAGE_WEBHOOKS = 1 << 29
    MANAGE_GUILD_EXPRESSION = 1 << 30
    USE_APPLICATION_COMMANDS = 1 << 31
    REQUEST_TO_SPEAK = 1 << 32
    MANAGE_EVENTS = 1 << 33
    MANAGE_THREADS = 1 << 34
    CREATE_PUBLIC_THREADS = 1 << 35
    CREATE_PRIVATE_THREADS = 1 << 36
    USE_EXTERNAL_STICKERS = 1 << 37
    SEND_MESSAGES_IN_THREADS = 1 << 38
    USE_EMBEDDED_ACTIVITES = 1 << 39
    MODERATE_MEMBERS = 1 << 40
    VIEW_CREATOR_MONETIZATION_ANALYTICS = 1 << 41
    USE_SOUNDBOARD = 1 << 42
    CREATE_GUILD_EXPRESSIONS = 1 << 43
    CREATE_EVENTS = 1 << 44
    USE_EXTERNAL_SOUNDS = 1 << 45
    SEND_VOICE_MESSAGES = 1 << 46

    def __init__(self, value: int, bot):
        self.bot = bot
        self.http = bot.http
        self.permissions = []
        self.raw_value = int(value)
        # print("here")
        self.calculate_permissions(self.raw_value)
        # print("ran")

    def calculate_permissions(self, perm_value: int):
        for key, value in vars(Permission).items():
            if type(value) != int:
                continue
            if (perm_value & value) == value:
                self.permissions.append({key: value})



import itertools
from time import perf_counter
from aioconsole import aprint
import asyncio
from .voice import Voice
from ..models import (
    Guild,
    Convert,
    User,
    Message,
    Member,
    MessageAck,
    PresenceUpdate,
    DMChannel,
    MessageAddReaction,
    MemberListUpdate,
    CallCreate
)

import ujson

class Handler:
    def __init__(self, bot) -> None:
        self.bot = bot

    async def handle_ready(self, data: dict):
        self._ready_data = data
        # with open("test.json", "a+") as f:
        #     ujson.dump(data, f, indent=4)

        self.bot.resume_url = data['resume_gateway_url'] + "?encoding=json&v=9&compress=zlib-stream"
        self.bot.session_id = data['session_id']
        
        guilds = data.get("guilds", [])
        private_channels = data.get("private_channels", [])
        users = data.get("users", [])
        relationships = data.get("relationships", [])
        merged_members = data.get("merged_members", [])
        for guild, channel, user, relation in itertools.zip_longest(
            guilds,
            private_channels,
            users,
            relationships,
        ):
            if guild is not None:
                self.bot.user.guilds.append(Guild(guild, self.bot))
            if channel is not None:
                chan = Convert(channel, self.bot)
                self.bot.user.private_channels.append(chan)
                self.bot.cached_channels[chan.id] = chan
            if user is not None:
                check_user = self.bot.fetch_user(user["id"])
                if check_user is None:
                    user = User(user, self.bot)
                    self.bot.cached_users[user.id] = user
                else:
                    check_user.partial_update(user)
            if relation is not None:
                check_user = self.bot.fetch_user(relation["id"])
                if check_user is None:
                    user = User(relation, self.bot)
                    self.bot.cached_users[user.id] = user
                
                    if relation["type"] == 1:
                        self.bot.user.friends.append(user)
                    if relation["type"] == 2:
                        self.bot.user.blocked.append(user)
                else:
                    check_user.partial_update(relation)

        await self.bot.emit("ready", perf_counter() - self.bot.startup)


    async def handle_ready_supplemental(self, data: dict):
        # Ok discord bad code
        # I have to use data from ready and this event to properly form payloads
        # Discord Bad CODE

        temp_members = {}
        temp_guilds = {}

        for guild, members, extra_members in itertools.zip_longest(
            data.get("guilds", []),
            self._ready_data.get("merged_members", []),
            data.get("merged_members", {}),
        ):
            # To get our own data
            if members is not None:
                index = self._ready_data.get("merged_members", []).index(members)
                temp_members.setdefault(str(index), []).extend(members)
            
                for member in members:
                    # print(member['roles'], member['user_id'])
                    check_user = self.bot.fetch_user(member['user_id'])
                    if check_user is None:
                        member = User(member, self.bot)
                        self.bot.cached_users[member.id] = member
                    else:
                        check_user.partial_update(member)
                
            # To get other data
            if extra_members is not None:
                index = data.get("merged_members", []).index(extra_members)
                temp_members.setdefault(str(index), []).extend(extra_members)

                for member in extra_members:

                    check_user = self.bot.fetch_user(member['user_id'])
                    if check_user is None:
                        member = User(member, self.bot)
                        self.bot.cached_users[member.id] = member
                    else:
                        check_user.partial_update(member)

            # DISCORD BAD
            # Add everything to a Guild
            # Discord bad
            if guild is not None:
                guilds: list = data.get("guilds", [])
                index = guilds.index(guild)
                temp_guilds.setdefault(str(index), []).append(guild)
                # print(temp_guilds[str(index)])
                # print(index, guild)

                check_guild = self.bot.fetch_guild(guild['id'])
                if check_guild is None:
                    guild = Guild(guild, self.bot)
                    self.bot.user.guilds.append(guild)
                    for member in temp_members[str(index)]:
                        check_user = guild.fetch_member(member['user_id'])
                        if check_user is None:
                            user = Member(member, self.bot)
                            guild.members.append(user)
                  
                        else:
                            check_user.partial_update(member)
                            
                            guild.members.append(check_user)
                  
                else:
                    check_guild.partial_update(guild)
                    for member in temp_members[str(index)]:
                        check_user = check_guild.fetch_member(member['user_id'])
                        if check_user is None:
                            user = Member(member, self.bot)
                            check_guild.members.append(user)
                        else:
                            check_user.partial_update(member)
                            check_guild.members.append(check_user)


        # DISCORD BAD
        merged_presences = data.get("merged_presences", {})
        guilds = merged_presences.get("guilds")
        for index, users in enumerate(guilds):
            temp_members.setdefault(str(index), []).extend(users)
            for user in users:
                # print(user['user_id'])
                check_user = self.bot.fetch_user(user['user_id'])
                if check_user is None:
                    user = User(user, self.bot)
                    self.bot.cached_users[user.id] = user
                else:
                    check_user.partial_update(user)
      
        friends = merged_presences.get("friends")
        for user in friends:
            check_user = self.bot.fetch_user(user['user_id'])
            if check_user is None:
                user = User(user, self.bot)
                self.bot.cached_users[user.id] = user
            else:
                check_user.partial_update(user)

        # DISCORD BAD x100000
        for index, indexed_guild in temp_guilds.items():
            members = temp_members[str(index)]
            guild = self.bot.fetch_guild(indexed_guild[0]['id'])
            if guild is not None:
                for member in members:
                    check_user = guild.fetch_member(member['user_id'])
        
                    if check_user is None:
                        member = Member(member, self.bot)
                        
                        guild.members.append(member)
                    else:
                        check_user.partial_update(member)
                        
                        guild.members.append(check_user)
            
        


        await self.bot.inbuilt_commands()
        await self.bot.emit("ready_supplemental")
        await asyncio.sleep(4)

        for guild in self.bot.user.guilds:
            try:
                if guild.member_count >= 1000:
                    n = 0
                    for channel in guild.channels:
                        if channel.type == 0:
                            n += 1
                            await self.bot.gateway.cache_guild(guild, guild.channels[0])
                        if n >= 3:
                            break
            except:
                continue
                    
    
    async def handle_message_create(self, data: dict):
        message = Message(data, self.bot)
        self.bot.cached_messages[message.id] = message
        if message.author.id not in self.bot.cached_users:
            self.bot.cached_users[message.author.id] = message.author
        
        # await aprint("Processing")
        await self.bot.process_commands(message)
        # await aprint("Processed")
        await self.bot.emit("message", message)

    async def handle_message_update(self, data: dict):
        message = Message(data, self.bot)
        self.bot.cached_messages[message.id] = message
        await self.bot.emit("message_update", message)

    async def handle_message_ack(self, data: dict):
        ack = MessageAck(data, self.bot)
        await self.bot.emit("message_ack", ack)

    async def handle_message_delete(self, data: dict):
        deleted_message = self.bot.fetch_message(data['id'])
        await self.bot.emit("message_delete", deleted_message)

    async def handle_message_reaction_add(self, data: dict):
        await self.bot.emit("message_reaction_add", MessageAddReaction(data, self.bot))
        

    async def handle_channel_create(self, data: dict):
        channel = Convert(data, self.bot)
        self.bot.cached_channels[channel.id] = channel

        if data.get("guild_id") is not None:
            guild = self.bot.fetch_guild(channel.guild_id)
            guild.channels.append(channel)
        else:
            self.bot.user.private_channels.append(channel)

        await self.bot.emit("channel_create", channel)

    async def handle_channel_delete(self, data: dict):
        deleted_channel = self.bot.get_channel(data['id'])
        await self.bot.emit("channel_delete", deleted_channel)
        del deleted_channel

    async def handle_thread_create(self, data: dict):
        # print(data)
        pass

    async def handle_thread_update(self, data: dict):
        # print(data)
        pass

    async def handle_thread_delete(self, data: dict):
        # print(data)
        pass

    async def handle_guild_create(self, data: dict):
        guild = Guild(data, self.bot)
        self.bot.user.guilds.append(guild)
        await self.bot.emit("guild_create", guild)

    async def handle_guild_delete(self, data: dict):
        guild = self.bot.fetch_guild(data['id'])
        await self.bot.emit("guild_delete", guild)
        del guild

    async def handle_guild_member_list_update(self, data: dict):
        await self.bot.emit("guild_member_list_update", MemberListUpdate(data, self.bot))

    async def handle_call_update(self, data: dict):
        pass

    async def handle_call_create(self, data: dict):
        await self.bot.emit("call_create", CallCreate(data, self.bot))

    async def handle_presence_update(self, data: dict):
        pres = PresenceUpdate(data, self.bot)
        if isinstance(pres.user, User):
            pres.user.partial_update(data)
        await self.bot.emit("presence_update",pres)


    async def handle_voice_state_update(self, data: dict):
        pass

    async def handle_voice_server_update(self, data: dict):
        self.bot.gateway.voice = Voice(bot=self.bot, voice_token=data['token'], endpoint=data['endpoint'], server_id=data.get("channel_id", data['guild_id']))
        await self.bot.emit("voice_server_update", self.bot.gateway.voice)

    async def handle_thread_list_sync(self, data: dict):
        pass

    async def handle_guild_member_chunk(self, data: dict):
        pass



from typing import Optional, Literal
import time

class Activity:
    def __init__(self) -> None:
        self.dict = {}


    def remove_null(self, dic: dict):
        new = {}
        for key, value in dic.items():
            if value == None:
                continue
            if isinstance(value, dict):
        
                value = self.remove_null(value)
          
            if value == {}:
                continue
            new[key] = value
        return new



    def Playing(
        self,
        application_id: str,
        name: str,
        details: Optional[str] = None,
        state: Optional[str] = None,
        start: Optional[int] = None,
        end: Optional[int] = None,
        large_image: Optional[str] = None,
        large_text: Optional[str] = None,
        small_image: Optional[str] = None,
        small_text: Optional[str] = None,
        buttons: list[dict[str, str]] = [],
        party_id: Optional[str] = None,
        party_size: Optional[list[int]] = None,
        match_secret: Optional[str] = None,
        join_secret: Optional[str] = None,
        spectate_secret: Optional[str] = None,
        instance: Optional[bool] = None
    ):
        self.dict['type'] = 0
        self.dict['application_id'] = application_id
        self.dict['name'] = name
        self.dict['details'] = details
        self.dict['state'] = state
        self.dict['timestamps'] = {"start": start, "end": end}
        self.dict['assets'] = {"large_image": large_image, "large_text": large_text, "small_image": small_image, "small_text": small_text}
        if len(buttons) > 0:
            if len(buttons) > 2:
                raise ValueError("You can only have 2 buttons")
            self.dict['buttons'] = []
            self.dict['metadata'] = []
            for button in buttons:
                self.dict['buttons'].append(button['label'])
                self.dict['metadata'].append(button['url'])
        self.dict['party'] = {"id": party_id, "size": party_size}
        self.dict['secrets'] = {"match": match_secret, "join": join_secret, "spectate": spectate_secret}

        self.dict['created_at'] = int(time.time())
        self.dict['instance'] = instance
        
        return self.remove_null(self.dict)


    def Streaming(
        self,
        application_id: str,
        name: str,
        url: Optional[str] = None,
        details: Optional[str] = None,
        state: Optional[str] = None,
        start: Optional[int] = None,
        end: Optional[int] = None,
        large_image: Optional[str] = None,
        large_text: Optional[str] = None,
        small_image: Optional[str] = None,
        small_text: Optional[str] = None,
        buttons: list[dict[str, str]] = [],
    ):
        self.dict['type'] = 1
        self.dict['application_id'] = application_id
        self.dict['name'] = name
        self.dict['url'] = url
        self.dict['details'] = details
        self.dict['state'] = state
        self.dict['timestamps'] = {"start": start, "end": end}
        self.dict['assets'] = {"large_image": large_image, "large_text": large_text, "small_image": small_image, "small_text": small_text}
        if len(buttons) > 0:
            if len(buttons) > 2:
                raise ValueError("You can only have 2 buttons")
            self.dict['buttons'] = []
            self.dict['metadata'] = []
            for button in buttons:
                self.dict['buttons'].append(button['label'])
                self.dict['metadata'].append(button['url'])
        self.dict['created_at'] = int(time.time())
     
        return self.remove_null(self.dict)


    def Listening(self,
        application_id: str,
        name: str,
        details: Optional[str] = None,
        state: Optional[str] = None,
        start: Optional[int] = None,
        end: Optional[int] = None,
        large_image: Optional[str] = None,
        large_text: Optional[str] = None,
        small_image: Optional[str] = None,
        small_text: Optional[str] = None,
        buttons: list[dict[str, str]] = [],
    ):
        self.dict['type'] = 2
        self.dict['application_id'] = application_id
        self.dict['name'] = name
        self.dict['details'] = details
        self.dict['state'] = state
        self.dict['timestamps'] = {"start": start, "end": end}
        self.dict['assets'] = {"large_image": large_image, "large_text": large_text, "small_image": small_image, "small_text": small_text}
        if len(buttons) > 0:
            if len(buttons) > 2:
                raise ValueError("You can only have 2 buttons")
            self.dict['buttons'] = []
            self.dict['metadata'] = []
            for button in buttons:
                self.dict['buttons'].append(button['label'])
                self.dict['metadata'].append(button['url'])
        self.dict['created_at'] = int(time.time())
 
        return self.remove_null(self.dict)
    
    def Watching(self,
        application_id: str,
        name: str,
        details: Optional[str] = None,
        state: Optional[str] = None,
        start: Optional[int] = None,
        end: Optional[int] = None,
        large_image: Optional[str] = None,
        large_text: Optional[str] = None,
        small_image: Optional[str] = None,
        small_text: Optional[str] = None,
        buttons: list[dict[str, str]] = [],
    ):
        self.dict['type'] = 3
        self.dict['application_id'] = application_id
        self.dict['name'] = name
        self.dict['details'] = details
        self.dict['state'] = state
        self.dict['timestamps'] = {"start": start, "end": end}
        self.dict['assets'] = {"large_image": large_image, "large_text": large_text, "small_image": small_image, "small_text": small_text}
        if len(buttons) > 0:
            if len(buttons) > 2:
                raise ValueError("You can only have 2 buttons")
            self.dict['buttons'] = []
            self.dict['metadata'] = []
            for button in buttons:
                self.dict['buttons'].append(button['label'])
                self.dict['metadata'].append(button['url'])
        self.dict['created_at'] = int(time.time())

        return self.remove_null(self.dict)

    def Custom(self, state: str, emoji: Optional[str] = None, id: Optional[str] = None, animated: Optional[bool] = None):
        self.dict['type'] = 4
        self.dict['state'] = state
        self.dict['emoji'] = {"name": emoji, "id": id, "animated": animated}

        self.remove_null(self.dict)
        return self.remove_null(self.dict)

    def Competing(self,
        application_id: str,
        name: str,
        details: Optional[str] = None,
        state: Optional[str] = None,
        start: Optional[int] = None,
        end: Optional[int] = None,
        large_image: Optional[str] = None,
        large_text: Optional[str] = None,
        small_image: Optional[str] = None,
        small_text: Optional[str] = None,
        buttons: list[dict[str, str]] = [],
    ):
        self.dict['type'] = 5
        self.dict['application_id'] = application_id
        self.dict['name'] = name
        self.dict['details'] = details
        self.dict['state'] = state
        self.dict['timestamps'] = {"start": start, "end": end}
        self.dict['assets'] = {"large_image": large_image, "large_text": large_text, "small_image": small_image, "small_text": small_text}
        if len(buttons) > 0:
            if len(buttons) > 2:
                raise ValueError("You can only have 2 buttons")
            self.dict['buttons'] = []
            self.dict['metadata'] = []
            for button in buttons:
                self.dict['buttons'].append(button['label'])
                self.dict['metadata'].append(button['url'])
        self.dict['created_at'] = int(time.time())
       
        return self.remove_null(self.dict)




from __future__ import annotations
from typing import TYPE_CHECKING, Optional
from zlib import decompressobj
import time
import asyncio
from .events import Handler
import websockets
from aioconsole import aprint
import ujson
from websockets.client import connect, ClientConnection
from .errors import ReconnectWebsocket
from websockets.exceptions import ConnectionClosed, ConnectionClosedError, ConnectionClosedOK
if TYPE_CHECKING:
    from ..bot import Bot
    from websockets import connection
    from ..models import Capabilities, Guild, Messageable
    from .voice import Voice


class Gateway:

    DISPATCH = 0
    HEARTBEAT = 1
    IDENTIFY = 2
    PRESENCE = 3
    VOICE_STATE = 4
    VOICE_PING = 5
    RESUME = 6
    RECONNECT = 7
    REQUEST_MEMBERS = 8
    INVALIDATE_SESSION = 9
    HELLO = 10
    HEARTBEAT_ACK = 11
    GUILD_SYNC = 12

    def __init__(self, bot: Bot, decompress: bool = True) -> None:
        self.decompress = decompress
        self.bot: Bot = bot
        self.capabilities: Capabilities = self.bot.capabilities
        self.handler: Handler = Handler(bot)
        self.token: Optional[str] = None
        self.zlib = decompressobj(15)
        self.zlib_suffix: bytes = b"\x00\x00\xff\xff"
        self.last_ack: float = 0
        self.last_send: float = 0
        self.latency: float = float("inf")
        self.ws = None
        self.alive = False
        self.URL = (
            "wss://gateway.discord.gg/?encoding=json&v=9&compress=zlib-stream"
            if self.decompress else 
            "wss://gateway.discord.gg/?encoding=json&v=9"
        )
        self.voice: Optional[Voice] = None
        self.subscriptions_data = {}

    async def linux_run(self, cmd):
        proc = await asyncio.create_subprocess_shell(
            cmd,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE
        )
        stout, stderr = await proc.communicate()




    async def send_json(self, payload: dict):
        if self.ws:
            try:
                await self.ws.send(ujson.dumps(payload))
            except ConnectionClosed as e:
                await self.handle_reconnect(e)
        

    async def load_async(self, item):
        loop = asyncio.get_event_loop()
        return await loop.run_in_executor(None, ujson.loads, item)

    async def recv_json(self):
        if self.ws:
            
            item = await self.ws.recv()
    
            if self.decompress:
                buffer = bytearray()
                try:
                    buffer.extend(item)
                except:
                    buffer = item
                if len(item) < 4 or item[-4:] != self.zlib_suffix:
                    return
                n = len(item)
                
                
                item = self.zlib.decompress(item)
                self.zlib.flush(n)
                
                
            item = await self.load_async(item)

            # await asyncio.sleep(1)

            if item:
                self.op = item["op"]
                data = item["d"]
                event = item["t"]
                self.seq = item["s"]

                if self.op == self.HELLO:
                    interval = data["heartbeat_interval"] / 1000.0
                    await self.identify()
                    asyncio.create_task(self.heartbeat(interval))

                elif self.op == self.HEARTBEAT_ACK:
                    self.heartbeat_ack()

                elif self.op == self.INVALIDATE_SESSION:
                    if data == False:
                        await self.close()
                        raise ReconnectWebsocket("Invalidated Session", resume=False, op=self.op)
                    else:
                        await self.close()
                        await asyncio.sleep(2)
                        await self.connect(self.bot.resume_url)
                        await self.resume()
                    

                elif self.op == self.RECONNECT:
                    # await aprint(f"Attempting reconnect???? {self.bot.user.username}")
                    # await self.linux_run(f"notify-send 'RECONNECT HAPPENING NOW CHECK CONSOLE {data} {op}'")
                    await self.close()
                    await asyncio.sleep(2)
                    await self.connect(self.bot.resume_url)
                    await self.resume()
                    

                elif self.op == self.DISPATCH:
                    if hasattr(self.handler, f"handle_{event.lower()}"):
                        method = getattr(
                            self.handler, f"handle_{event.lower()}")
                        asyncio.create_task(method(data))

    async def connect(self, url: str):
        self.ws = await connect(
            url, origin="https://discord.com", max_size=None,
            extra_headers={"user_agent": "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/118.0.0.0"},
            read_limit=1000000, max_queue=100, write_limit=1000000,
        )
        self.alive = True
        self.zlib = decompressobj(15)

    async def handle_reconnect(self, e: ConnectionClosed):
        # await aprint(f"Closing because fail. Attempting reconnect {self.bot.user.username}\n{e}")
        if e.rcvd is not None:
            if e.rcvd.code in [1000, 1001, 1011, 4000, 4001, 4002, 4003, 4005, 4007, 4008, 4009]:
                # await aprint(f"RECEIVE: {e.rcvd.code}  --- {e.rcvd.reason}")
                await self.close()
                await asyncio.sleep(2)
                await self.connect(self.bot.resume_url)
                await self.resume()
                
        if e.sent is not None:
            if e.sent.code in [1000, 1001, 1011, 4000, 4001, 4002, 4003, 4005, 4007, 4008, 4009]:
                # await aprint(f"SENT: {e.sent.code} --- {e.sent.reason}")
                await self.close()
                await asyncio.sleep(2)
                await self.connect(self.bot.resume_url)
                await self.resume()

        
        # await aprint("Raising error")
        await self.close()
        raise ReconnectWebsocket("Unknown", resume=False, op=False)

    async def start(self, token: str, resume: bool = False):
        if not resume:
            await self.connect(self.URL)
            
            self.token = token
            while self.alive:
                try:

                    await self.recv_json()

                except ConnectionClosed as e:
                    await self.handle_reconnect(e)

                



    async def resume(self):
        # await aprint(self.seq, self.bot.session_id)
        await self.send_json({
            "op": self.RESUME,
            "d": {"token": self.token, "session_id": self.bot.session_id, "seq":self.seq},
        })
        
    async def cache_guild(self, guild: Guild, channel):
        payload = {
            "op": 14,
            "d": {
                "guild_id": guild.id,
                "typing": True,
                "threads": False,
                "activities": True,
                "members": [],
                "channels": {
                    str(channel.id): [
                        [
                            0,
                            99
                        ]
                    ]
                }
            }
        }
        await self.send_json(payload)

    async def close(self):
        """This function closes the websocket
        """
        self.alive = False
        await self.ws.close()

    async def identify(self):
        payload = {
            "op": 2,
            "d": {
                "capabilities": self.capabilities.value,
                "token": self.token,
                "client_state": {
                    "guild_versions": {},
                    "api_code_version": 0,
                    "highest_last_message_id": "0",
                    "initial_guild_id": None,
                    "private_channels_version": "0",
                    "read_state_version": 0,
                    "user_guild_settings_version": -1,
                    "user_settings_version": -1,
                },
                "compress": False,
                "presence": {
                    "activities": [],
                    "afk": False,
                    "since": 0,
                    "status": "online",
                },
                "properties": {
                    "os": "Linux",
                    "browser": "Chrome",
                    "device": "",
                    "system_locale": "en-GB",
                    "browser_user_agent": "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/118.0.0.0 Safari/537.36",
                    "browser_version": "118.0.0.0",
                    "os_version": "",
                    "referrer": "",
                    "referring_domain": "",
                    "referrer_current": "",
                    "referring_domain_current": "",
                    "release_channel": "canary",
                    "client_build_number": 259975,
                    "client_event_source": None
                },

            },
        }
        await self.send_json(payload)



    async def heartbeat(self, interval: int):
        heartbeat_json = {"op": 1, "d": time.time()}
        while True:
            await asyncio.sleep(interval)
            await self.send_json(heartbeat_json)
            self.last_send = time.perf_counter()

    def heartbeat_ack(self):
        self.last_ack = time.perf_counter()
        self.latency = self.last_ack - self.last_send

    def roundup(self, n):
        import math
        return int(math.ceil(n / 100.0)) * 100
    
    def chunks(self, lst, n):
        for i in range(0, len(lst), 1):
            if len(lst[: i + 1]) > 3:
                for i in range(i, len(lst), n):
                    yield lst[i : i + n]
                break
            yield lst[: i + 1]

    def correct_channels(self, guild: Guild):
        roles = guild.me.roles
        
        channels = []
        for channel in guild.channels:
            if len(channel.permission_overwrites) > 0:
                for overwrite in channel.permission_overwrites:
                    if (overwrite.id == guild.id) or (overwrite.id in [role.id for role in roles]):
                        for permission in overwrite.allow.permissions:
                            for name, value in permission.items():
                                if name == "VIEW_CHANNEL":
                                    if value:
                                        
                                        channels.append(channel)
                                        break
  
    
                    for permission in overwrite.deny.permissions:
                        for name, value in permission.items():
                            if name == "VIEW_CHANNEL":
                                break
                        else:
                            
                            channels.append(channel)
                            break

        return list(set(channels))
    
    def get_ranges(self, amount: int):
        ranges = []
        for i in range(0, amount, 100):
            ranges.append(
                [i, self.roundup(i + (amount - i)) - 1]
            ) if i + 99 > amount else ranges.append([i, i + 99])
        return ranges
    
    async def subscriptions(self, guild: Guild, channels: Optional[list[Messageable]] = None):
       # In Progres...
       # Basically discord no longer uses op 14, uses this now
        if guild.member_count is not None:
            ranges = self.get_ranges(guild.member_count)
        if channels is None:
            channels = self.correct_channels(guild)
        self.subscriptions_data[guild.id] = {"channels": {}}
        self.subscriptions_data[guild.id]["channels"].update({channels[0].id: [0, 99]})

        payload = {
            "op": 37, 
            "d": {
                "subscriptions": self.subscriptions_data
            }
        }
        await self.send_json(payload)


    async def chunk_members(self, guild: Guild):
        channels = self.correct_channels(guild)
        channels = channels[:5]
        ranges = []

        if guild.member_count is not None:
            for i in range(0, guild.member_count, 100):
                ranges.append(
                    [i, self.roundup(i + (guild.member_count - i)) - 1]
                ) if i + 99 > guild.member_count else ranges.append([i, i + 99])
  
            
        for item in self.chunks(ranges, 3):

            queries = {}
            payload = payload = {
                "op": 14,
                "d": {
                    "guild_id": guild.id,
                    "typing": True,
                    "threads": False,
                    "activities": True,
                }
            }
            data = payload['d']

            # For now
            for channel in channels:
                queries[channel.id] = item

            data['channel'] = queries
            
            await self.send_json(payload)
            # print(payload)
            await asyncio.sleep(2.0)
        

    async def call(self, channel: str, guild: Optional[str] = None):
        payload = {
            "op": 4,
            "d": {
                "guild_id": guild,
                "channel_id": channel,
       
                "self_mute": False,
                "self_deaf": False,
                "self_video": False,
            },
        }
        await self.send_json(payload)


    async def video_call(self, channel: str, guild: Optional[str] = None):
        payload = {
            "op": 4,
            "d": {
                "guild_id": guild,
                "channel_id": channel,
    
                "self_mute": False,
                "self_deaf": False,
                "self_video": True,
            },
        }
        await self.send_json(payload)

    async def leave_call(self):
        payload = {
            "op": 4,
            "d": {
                "guild_id": None,
                "channel_id": None,
                "self_mute": False,
                "self_deaf": False,
                "self_video": False,
            },
        }
        await self.send_json(payload)



from __future__ import annotations
import asyncio
import itertools
from typing import TYPE_CHECKING, Optional, Literal, AsyncGenerator
from .assets import Asset
from .channels import Convert, Messageable
from .message import Message
from .users import Member
from .permissions import Permission

if TYPE_CHECKING:
    from ..bot import Bot


class Guild:
    def __init__(self, payload: dict, bot: Bot):
        self.bot = bot
        self.http = bot.http
        self.update(payload)

    @property
    def me(self) -> Member:
        return self.members[0]

    def update(self, payload: dict):
        self.members: list[Member] = []
        self.channels: list[Messageable] = []
        self.emojis: list[Emoji] = []
        self.stickers: list[Sticker] = []
        self.roles: list[Role] = []
        # MUH OPTIMISATIONS: Zip Longest very cool bro
        self.id: Optional[str] = payload.get("id")
        for emoji, sticker, role, channel, member in itertools.zip_longest(
            payload.get("emojis", []),
            payload.get("stickers", []),
            payload.get("roles", []),
            payload.get("channels", []),
            payload.get("members", [])
        ):
            if emoji is not None:
                self.emojis.append(Emoji(emoji, self.bot))

            if sticker is not None:
                self.stickers.append(Sticker(sticker, self.bot))

            if role is not None:
                role.update({"guild_id": self.id})
                role = Role(role, self.bot)

                self.roles.append(role)

            if channel is not None:
                channel.update({"guild_id": self.id})
                chan = Convert(channel, self.bot)

                self.channels.append(chan)
             
                self.bot.cached_channels[chan.id] = chan

            if member is not None:
                member.update({"guild_id": self.id})
                member = Member(member, self.bot)
                self.members.append(member)
                

        self.member_count = payload.get("member_count")
        self.embedded_activities = payload.get("embedded_activities", [])
        self.voice_states = payload.get("voice_state", [])
        self.lazy = payload.get("lazy")
        self.large = payload.get("large")
        self.joined_at = payload.get("joined_at")
        properties: Optional[dict] = payload.get("properties")
        if properties:
            self.id: Optional[str] = properties.get("id")
            self.owner_id: Optional[str] = (
                properties["owner_id"]
                if properties.get("owner_id") is not None
                else None
            )
            self.premium_tier: Optional[int] = properties.get("premium_tier")
            self.splash: Optional[Asset] = (
                Asset(self.id, properties["splash"])
                if properties.get("splash") is not None
                else None
            )
            self.nsfw_level: Optional[str] = (
                properties["nsfw_level"]
                if properties.get("nsfw_level") is not None
                else None
            )
            self.application_id: Optional[str] = (
                properties["application_id"]
                if properties.get("application_id") is not None
                else None
            )
            self.system_channel_flags: Optional[int] = properties.get(
                "system_channel_flags"
            )
            self.inventory_settings: Optional[str] = properties.get(
                "inventory_settings"
            )
            self.default_message_notifications: Optional[int] = properties.get(
                "default_message_notifications"
            )
            self.hub_type: Optional[int] = properties.get("hub_type")
            self.afk_channel: Optional[str] = properties.get("afk_channel")
            self.incidents_data: Optional[int] = properties.get(
                "incidents_data")
            self.discovery_splash: Optional[Asset] = (
                Asset(self.id, properties["discovery_splash"])
                if properties.get("discovery_splash") is not None
                else None
            )
            self.preferred_locale: Optional[str] = properties.get(
                "preferred_locale")
            
            self.icon: Optional[Asset] = (
                Asset(self.id, properties["icon"]).from_icon()
                if properties.get("icon") is not None
                else None
            )
            self.latest_onboarding_question_id: Optional[str] = properties.get(
                "latest_onboarding_question_id"
            )
            self.explicit_content_filter: Optional[int] = properties.get(
                "explicit_content_filter"
            )
            self.description: Optional[str] = properties.get("description")
            self.afk_timeout: Optional[int] = properties.get("afk_timeout")
            self.max_video_channel_users: Optional[int] = properties.get(
                "max_video_channel_users"
            )
            self.nsfw: Optional[bool] = properties.get("nsfw")
            self.system_channel_id: Optional[str] = properties.get(
                "system_channel_id")
            self.rules_channel_id: Optional[str] = properties.get(
                "rules_channel_id")
            self.max_stage_video_channel_users: Optional[int] = properties.get(
                "max_stage_video_channel_users"
            )
            self.banner: Optional[Asset] = (
                Asset(self.id, properties["banner"])
                if properties.get("banner") is not None
                else None
            )
            self.public_updates_channel_id: Optional[int] = properties.get(
                "public_updates_channel_id"
            )
            self.mfa_level: Optional[int] = properties.get("mfa_level")
            self.features: Optional[list[str]] = properties.get("features")
            self.max_members: Optional[int] = properties.get("max_members")
            self.name: Optional[str] = properties.get("name")
            self.safety_alerts_channel_id: Optional[int] = properties.get("safety_alerts_channel_id")
            self.premium_progress_bar_enabled: Optional[bool] = properties.get("premium_progress_bar_enabled")
            self.verification_level: Optional[int] = properties.get("verification_level")
            self.home_header: Optional[str] = properties.get("home_header")
            self.vanity_url_code: Optional[str] = properties.get("vanity_url_code")

    def partial_update(self, payload: dict):
        for key, value in payload.items():
            if hasattr(self, key):
                if key == "properties":
                    for key, value in payload.items():
                        if key == "banner":
                            setattr(self, key, (
                                Asset(self.id, payload["banner"]).from_avatar()
                                if payload.get("banner") is not None and self.id is not None
                                else None
                            ))
                        elif key == "icon":
                      
                            setattr(self, key, (
                                Asset(self.id, payload["banner"]).from_avatar()
                                if payload.get("banner") is not None and self.id is not None
                                else None
                            ))
                        elif key == "bot":
                            setattr(self, "is_bot", value)

                if key == "banner":
                    setattr(self, key, (
                        Asset(self.id, payload["banner"]).from_avatar()
                        if payload.get("banner") is not None and self.id is not None
                        else None
                    ))
                elif key == "icon":
                    setattr(self, key, (
                        Asset(self.id, payload["banner"]).from_avatar()
                        if payload.get("banner") is not None and self.id is not None
                        else None
                    ))
                elif key == "bot":
                    setattr(self, "is_bot", value)

                else:
                    setattr(self, key, value)

    def fetch_member(self, user_id: str):
        for member in self.members:
            if member.id == user_id:
                return member
            
    async def search_base(
        self,
        content: Optional[str] = None,
        has: Optional[Literal["link", "embed", "file", "video", "image", "sound"]] = None,
        from_: Optional[str] = None,
        mentions: Optional[str] = None,
        max: Optional[int] = None,
        min: Optional[int] = None,
        channel_id: Optional[str] = None,
        offset: Optional[int] = None
    ) -> Optional[list[Message]]:
        params = {
            "content": content,
            "has": has,
            "from": from_,
            "mentions": mentions,
            "max": max,
            "min": min,
            "channel_id": channel_id,
            "offset": offset
        }
        params = {k: v for k, v in params.items() if v is not None}
        json = await self.http.request(
            "GET", f"/guilds/{self.id}/messages/search", params=params
        )
        if json is not None:
            for messages in json['messages']:
                for message in messages:
                    self.bot.cached_messages[message.id] = Message(message, self.bot)
            return [Message(message, self.bot) for messages in json['messages'] for message in messages]

    async def search(
        self,
        content: Optional[str] = None,
        has: Optional[Literal["link", "embed", "file", "video", "image", "sound"]] = None,
        from_: Optional[str] = None,
        mentions: Optional[str] = None,
        max: Optional[int] = None,
        min: Optional[int] = None,
        channel_id: Optional[str] = None,
        amount: int = 25
    ) -> list[Message]:
        n = 0
        msgs = []
        while n < amount:
            msg = await self.search_base(
                content=content,
                has=has,
                from_=from_,
                mentions=mentions,
                max=max,
                min=min,
                channel_id=channel_id,
                offset=n
            )
            if msg is not None:
                if len(msg) == 0:
                    break
                n += len(msg)
                msgs.extend(msg)

        return msgs

    async def get_members(self):
        asyncio.create_task(self.bot.gateway.chunk_members(self))

    async def delete(self):
        await self.http.request(
            "POST", f"/guilds/{self.id}/delete"
        )
    
    async def leave(self, lurking: bool = False):
        await self.http.request(
            "DELETE", f"/users/@me/guilds/{self.id}", json={"lurking": lurking}
        )


    async def create_role(self, name: str, color: int = 0, permissions: int = 0):
        json = await self.http.request(
            "POST", f"/guilds/{self.id}/roles",
            json={
                "name": name,
                "color": color,
                "permissions": permissions
            }
        )
        if json is not None:
            return Role(json, self.bot)

    async def create_channel(self, name: str, type: int, permission_overwrites: list[dict] = []) -> Optional[Messageable]:
        json = await self.http.request(
            "POST",
            f"/guilds/{self.id}/channels",
            json={"name": name, "type": type, "permission_overwrites": permission_overwrites}
        )
        return Convert(json, self.bot) or None
    
    async def create_text_channel(self, name: str, permission_overwrites: list[dict] = []) -> Optional[Messageable]:
        json = await self.http.request(
            "POST",
            f"/guilds/{self.id}/channels",
            json={"name": name, "type": 0, "permission_overwrites": permission_overwrites}
        )
        return Convert(json, self.bot) or None
    
    async def create_voice_channel(self, name: str, permission_overwrites: list[dict] = []) -> Optional[Messageable]:
        json = await self.http.request(
            "POST",
            f"/guilds/{self.id}/channels",
            json={"name": name, "type": 2, "permission_overwrites": permission_overwrites}
        )
        return Convert(json, self.bot) or None
    
    async def ban_list(self):
        
        json = await self.http.request(
            "GET",
            f"/guilds/{self.id}/bans?limit=1000",
        )
        json = json if json is not None else []
        return [Member(data['user'], self.bot) for data in json]

class Emoji:
    def __init__(self, payload: dict, bot: Bot):
        self.bot = bot
        self.http = bot.http
        self.update(payload)

    def update(self, payload: dict):
        self.roles = payload.get("roles")
        self.name = payload.get("name")
        self.require_colons = payload.get("require_colons")
        self.managed = payload.get("managed")
        self.id = payload["id"]
        self.available = payload.get("available")
        self.animated = payload.get("animated")


class Role:
    def __init__(self, payload: dict, bot: Bot):
        self.bot = bot
        self.http = bot.http
        self.update(payload)

    def update(self, payload: dict):
        self.unicode_emoji = payload.get("unicode_emoji")
        self.position = payload.get("position")
        self.permissions = Permission(payload['permissions'], self.bot) if payload.get("permissions") is not None else None
        self.name = payload.get("name")
        self.mentionable = payload.get("mentionable")
        self.managed = payload.get("managed")
        self.id = payload["id"]
        self.guild_id = payload.get("guild_id")
        self.icon = payload.get("icon")
        self.hoist = payload.get("hoist")
        self.flags = payload.get("flags")
        self.color = payload.get("color")


class Sticker:
    def __init__(self, payload: dict, bot: Bot):
        self.bot = bot
        self.http = bot.http
        self.update(payload)

    def update(self, payload: dict):
        self.type = payload.get("type")
        self.tags = payload.get("tags")
        self.name = payload.get("name")
        self.id = payload.get("id")
        self.guild_id = payload.get("guild_id")
        self.format_type = payload.get("format_type")
        self.description = payload.get("description")
        self.available = payload.get("available")
        self.asset = payload.get("asset")



from __future__ import annotations
from typing import Optional, TYPE_CHECKING
from .users import User
import asyncio

if TYPE_CHECKING:
    from ..bot import Bot
    from .channels import Messageable
    from .guild import Guild

class Message:
    def __init__(self, data: dict, bot: Bot):
        self.bot = bot
        self.http = bot.http
        self.update(data)

    @property
    def link(self):
        if self.guild_id is not None:
            link = f"https://discord.com/channels/{self.guild_id}/{self.channel_id}/{self.id}"
        else:
            link = f"https://discord.com/channels/@me/{self.channel_id}/{self.id}"
        return link
    
    @property
    def guild(self):
        return self.bot.fetch_guild(self.guild_id)
    
    @property
    def channel(self):
        return self.bot.get_channel(self.channel_id)

    def update(self, payload: dict):
        self.id: Optional[str] = payload.get("id")
        self.content: Optional[str] = payload.get("content")
        if self.content is not None:
            self.content = self.content.replace("\x00", "")
        self.type: int = payload.get("type", 0)
        self.tts: bool = payload.get("tts", False)
        self.timestamp: Optional[int] = payload.get("timestamp")
        self.replied_message: Optional[Message] = payload.get("referenced_message")
        self.pinned: Optional[bool] = payload.get("pinned")
        self.nonce: Optional[int] = payload.get("nonce")
        self.mentions: Optional[dict] = payload.get("mentions")
        self.channel_id: Optional[str] = payload.get("channel_id", "")
        self.guild_id: Optional[str] = payload.get("guild_id")
        # if payload.get("author") is None:
        #     print(payload)
        self.author: Optional[User] = (
            User(payload['author'], self.bot)
            if payload.get("author") is not None
            else None
        )
        if self.author is not None:
            self.bot.cached_users[self.author.id] = self.author
        # we will fix later 
        # self.member = Member(payload.get("member"), self.bot)
        self.flags: int = payload.get("flags", 0)
        # Create associated classes with these
        self.embeds = payload.get("embeds")
        self.components = payload.get("components")
        self.attachments = payload.get("attachments")

    async def delete(self):
        await self.http.request(
            "DELETE", f"/channels/{self.channel_id}/messages/{self.id}"
        )

    async def reply(self, content: str, files: Optional[list[str]] = None, delete_after: Optional[int] = None, tts: bool = False) -> Optional[Message]:
        data = {
            "mobile_network_type":"unknown",
            "content":content,
            "tts":tts,
            "message_reference":{
                "channel_id":self.channel_id,
                "message_id":self.id
            },
            "allowed_mentions":{"parse":["users","roles","everyone"],"replied_user":True},"flags":0
        }
        if files is not None and len(files) > 0:
            uploaded_files = await self.channel.upload_image(files)
            data.update({"files": uploaded_files})

        json = await self.http.request(
            "POST", f"/channels/{self.channel_id}/messages",
            json=data
        )
        if json is not None:
            msg = Message(json, self.bot)
            if delete_after is not None:
                await asyncio.create_task(self.channel.delayed_delete(msg, delete_after))
            return msg
      

    async def edit(self, content: str, files: Optional[list[str]] = None) -> Optional[Message]:
        data: dict[str, int | str] = {"content": content}

        if files is not None:
            uploaded_files = await self.channel.upload_image(files)
            data.update({"files": uploaded_files})

        json = await self.http.request(
            "PATCH", f"/channels/{self.channel_id}/messages/{self.id}", 
            json=data
        )
        if json is not None:
            return Message(json, self.bot)
        
    async def add_reaction(self, emoji: str):
        await self.http.request(
            "PUT", f"/channels/{self.channel_id}/messages/{self.id}/reactions/{self.http.encode_emoji(emoji)}/%40me?location=Message&type=0"
        )

        
class MessageAck:
    def __init__(self, payload: dict, bot) -> None:
        self.bot = bot
        self.http = bot.http
        self.update(payload)

    def update(self, payload: dict):
        self.channel_id: str = payload['channel_id']
        self.channel = self.bot.get_channel(self.channel_id)
        self.flags: Optional[int] = payload.get("flags")
        self.last_viewed = payload['last_viewed']
        self.message_id = payload['message_id']
        self.message = self.bot.fetch_message(self.message_id)
        self.version = payload['version']

class EmbedField:
    def __init__(self, payload: dict, bot):
        self.bot = bot
        self.http = bot.http
        self.update(payload)

    def update(self, payload: dict):
        self.name = payload.get("name")
        self.value = payload.get("value")
        self.inline = payload.get("inline")

class EmbedThumbnail:
    def __init__(self, payload: dict, bot):
        self.bot = bot
        self.http = bot.http
        self.update(payload)

    def update(self, payload: dict):
        self.url = payload['url']
        self.proxy_url = payload.get("proxy_url")
        self.height = payload.get("height")
        self.width = payload.get("width")

class EmbedVideo:
    def __init__(self, payload: dict, bot):
        self.bot = bot
        self.http = bot.http
        self.update(payload)

    def update(self, payload: dict):
        self.url = payload['url']
        self.proxy_url = payload.get("proxy_url")
        self.height = payload.get("height")
        self.width = payload.get("width")

class EmbedImage:
    def __init__(self, payload: dict, bot):
        self.bot = bot
        self.http = bot.http
        self.update(payload)

    def update(self, payload: dict):
        self.url = payload['url']
        self.proxy_url = payload.get("proxy_url")
        self.height = payload.get("height")
        self.width = payload.get("width")

class EmbedProvider:
    def __init__(self, payload: dict, bot):
        self.bot = bot
        self.http = bot.http
        self.update(payload)

    def update(self, payload: dict):
        self.name = payload.get("name")
        self.url = payload.get("url")

class EmbedAuthor:
    def __init__(self, payload: dict, bot) -> None:
        self.bot = bot
        self.http = bot.http
        self.update(payload)
    
    def update(self, payload: dict):
        self.name = payload.get("name")
        self.url = payload.get("url")
        self.icon_url = payload.get("icon_url")
        self.proxy_icon_url = payload.get("proxy_icon_url")

class EmbedFooter:
    def __init__(self, payload: dict, bot) -> None:
        self.bot = bot
        self.http = bot.http
        self.update(payload)

    def update(self, payload: dict):
        self.text = payload.get("text")
        self.icon_url = payload.get("icon_url")
        self.proxy_icon_url = payload.get("proxy_icon_url")


class Embed:
    def __init__(self, payload: dict, bot):
        self.bot = bot
        self.http = bot.http
        self.update(payload)

    def update(self, payload: dict):
        self.fields: list[EmbedField] = [EmbedField(field, self.bot) for field in payload.get("fields", [])]
        self.title: Optional[str] = payload.get("title")
        self.description: Optional[str] = payload.get("description")
        self.type: Optional[str] = payload.get("type")
        self.url: Optional[str] = payload.get("url")
        self.timestamp: Optional[int] = payload.get("timestamp")
        self.color: Optional[int] = payload.get("color")
        self.footer: Optional[str] = payload.get("footer")
        self.image: Optional[dict] = payload.get("image")
        self.thumbnail: Optional[dict] = payload.get("thumbnail")
        self.video: Optional[dict] = payload.get("video")
        self.provider: Optional[dict] = payload.get("provider")
        self.author: Optional[str] = payload.get("author")





from __future__ import annotations
from typing import TYPE_CHECKING, Optional

if TYPE_CHECKING:
    from ..api.http import HttpClient
    from ..bot import Bot

class Event_Session:
    def __init__(self, data: dict, bot: Bot, http: HttpClient):
        self.bot = bot
        self.http = http
        self._update(data)

    def _update(self, data):
        self.id = data.get("session_id")
        self.status = data.get("status")
        info = data.get("client_info")
        self.os = info.get("os")
        self.platform = info.get("client")
        self.version = info.get("version")
        if self.id == "all":
            self.type = "presence"
        elif (self.platform == "mobile") and (self.os == "other"):
            self.type = "selfcord"
        else:
            self.type = "user"


class Session:
    def __init__(self, data: dict, bot: Bot, http: HttpClient):
        self.bot: Bot = bot
        self.http: HttpClient = http
        self._update(data)

    def _update(self, data):
        self.last_used: Optional[str] = data.get("approx_last_used_time")
        self.hash: Optional[str] = data.get("id_hash")
        info = data.get("client_info")
        self.location: Optional[str] = info.get("location")
        self.os: Optional[str] = info.get("os")
        self.platform: Optional[str] = info.get("platform")
        
       
    async def remove(self, password):
        """
        Method to logout of a particular session

        Args:
            password: Your password
        """
        await self.http.request("post", "/auth/sessions/logout", json={"session_id_hashes": [self.hash], "password": password})
        del self



from pathlib import Path
import sys
from setuptools import find_packages, setup

this_directory = Path(__file__).parent
if __name__ == "__main__":
    this_directory = Path(__file__).parent
    long_description = (this_directory / "README.md").read_text()
    setup(
        name="selfcord-fix.py",
        packages=find_packages(
            include=[
                "selfcord",
                "selfcord.api",
                "selfcord.utils",
                "selfcord.models",
                "selfcord.api.voice",
            ]
        ),
        version="1.0.3",
        description="A Discord API wrapper designed for selfbots!",
        readme="README.md",
        author="Shell",
        extras_require={
            "voice": ["pynacl==1.5.0", "opuslib==3.0.1"],
            "linux": ["uvloop==0.17.0"],
            "windows": ["winloop==0.1.0"]
        },
        license="MIT",
        # Remove the conditional if/else block and use environment markers
        install_requires=[
            "aiohttp==3.8.5",
            "aioconsole==0.3.3",
            "websockets==10.1",
            "ujson==5.7.0",
            "aiofiles==0.8.0",
            "requests==2.31.0",
            
            # Use markers to ensure only the correct library is installed
            "uvloop==0.17.0; sys_platform != 'win32'",  # Installed on Linux (Docker) and macOS
            "winloop==0.1.0; sys_platform == 'win32'"   # Installed ONLY on Windows
        ],
        setup_requires=["pytest-runner"],
        tests_require=["pytest"],
        test_suite="tests",
        keywords=["selfbot", "discord", "discordapi", "discordwrapper"],
        long_description=long_description,
        url="https://github.com/OmegaDevStudio/Selfcord",
        long_description_content_type="text/markdown",
    )



# from twocaptcha import TwoCaptcha

# # iirc doesn't work because i forgo
# # Use https://2captcha.com/?from=15127673
# class CaptchaSolver:
#     def __init__(self, api_key: str,
#         site_key: str = "4c672d35-0701-42b2-88c3-78380b0db560",
#         site_url: str = "discord.com"
#     ):
#         self.two_captcha = TwoCaptcha(apiKey=api_key)
#         self.site_key = site_key
#         self.site_url = site_url

#     async def _get_captcha_solution(self):
#         try:
#             solution = await self.two_captcha.hcaptcha(sitekey=self.site_key, url=self.site_url)
#             return solution
#         except Exception as e:
#             print(f"Failed to get captcha solution {e}")  # Replace this with your logger
#             return None

#     async def solve_captcha(self):
#         return await self._get_captcha_solution()



from __future__ import annotations

import ast
from time import perf_counter
import asyncio
import contextlib
import importlib
import inspect
import io
import os
import random
import sys
import time
import urllib

from collections import defaultdict
from traceback import format_exception
from typing import TYPE_CHECKING, Optional, Literal
from urllib.parse import urlparse

import aiofiles
import aiohttp
from aioconsole import aexec, aprint

from selfcord.models.sessions import Session

from .api import Gateway, HttpClient, ReconnectWebsocket
from .models import (
    Client, DMChannel, GroupChannel, Guild,
    Message, TextChannel, User, VoiceChannel,
    Capabilities, Convert, Messageable,
    Role, Callable
)

from .utils import (
    Command, CommandCollection, Context, Event, Extension,
    ExtensionCollection, logging
)
from .utils.logging import handler
import sys


if TYPE_CHECKING:
    from .api import Voice

if sys.platform == "linux":
    import uvloop
    uvloop.install()

if sys.platform.lower().startswith("win"):
    import winloop # type: ignore
    winloop.install()




class Bot:
    """Bot instance as entry point to interact with the bot

    Args:
        debug (bool): Whether to start the bot in debug mode, defaults to False.
        prefixes (list[str]): Prefixes for the bot, defaults to s!.
        inbuilt_help (bool): Whether the inbuilt help command should be enabled, defaults to True.
        userbot (bool): Whether the bot should be a userbot rather than selfbot, defaults to False.
        eval (bool): Whether to have the eval command as default, defaults to False.
    """

    def __init__(
        self,
        prefixes: list[str] = ["s!"],
        inbuilt_help: bool = True,
        userbot: bool = False,
        token_leader: Optional[str] = None,
        eval: bool = False,
        decompress: bool = True,
    	password: Optional[str] = None
    ) -> None:
        self.inbuilt_help: bool = inbuilt_help
        self.token: str
        self.http: HttpClient = HttpClient(self)
        self.t1: float = time.perf_counter()
        self.session_id: str
        self.resume_url: str
        self.capabilities: Capabilities = Capabilities.default()
        self._events = defaultdict(list)
        self.commands = CommandCollection()
        self.prefixes: list[str] = (
            prefixes if isinstance(prefixes, list) else [prefixes]
        )
        self.extensions = ExtensionCollection()
        self.user: Client
        self.eval: bool = eval
        self.token_leader = token_leader
        self.bots = []
        if self.token_leader is not None:
            self.userbot: bool = True
        else:
            self.userbot: bool = userbot
        self.password: str = password
        self.cached_users: dict[str, User] = {}
        self.cached_channels: dict[str, Messageable] = {}
        self.cached_messages: dict[str, Message] = {}
        self.gateway: Gateway = Gateway(self, decompress)
        self.startup = perf_counter()
        self.semaphore = asyncio.BoundedSemaphore(1)

    if os.name == "nt":
        asyncio.set_event_loop_policy(asyncio.WindowsProactorEventLoopPolicy())


        
    async def runner(self, token: str, multi_token: bool = False):
        data = await self.http.static_login(token)
        self.token = token
        if data is not None:
            
            self.user = Client(data, self)
       
            while True:
                try:
                    await self.gateway.start(token)
                except ReconnectWebsocket as e:
                    await self.gateway.close()
  


                

    def run(self, token: str):
        """Used to start connection to gateway as well as gather user information

        Args:
            token (str): _description_
        """
        self.token = token

        try: 
            asyncio.run(self.runner(token))
        except Exception as e:
            self.gateway.alive = False
            raise e

    @property
    def latency(self):
        """Latency of heartbeat ack, gateway latency essentially"""
        return self.gateway.latency

    # For events
    async def inbuilt_commands(self):
        """
        I call this on bot initialisation, it's the inbuilt help command
        """
        if self.inbuilt_help:
            @self.cmd("The help command!", aliases=["test"])
            async def help(ctx, cat=None):
                """The help command, dedicated to viewing all commands, extensions and information regarding commands."""
                if cat is None:
                    msg = f"```ini\n"
                    msg += f"[ {self.user} ]\nType <prefix>help <ext_name> to view commands relating to a specific extension. Type <prefix>help <cmd_name> to view information regarding a command.\n[ .Prefixes ] : {self.prefixes}\n\n"
                    msg += f"[ .Commands ]\n"
                    for command in self.commands:
                        msg += f". {command.name}: {command.description}\n"
                    msg += "\n[ .Extensions ]\n"
                    for ext in self.extensions:
                        msg += f"[ {ext.name} ] : [ {ext.description} ]\n"

                    msg += "```"
                    return await ctx.reply(f"{msg}")

                else:
                    name = cat.lower()
                    for ext in self.extensions:
                        if name == ext.name.lower():
                            msg = f"```ini\n[ {self.user.username} Selfbot ]\n"
                            msg += f"[ {self.user} ]\n\nType <prefix>help <ext_name> to view commands relating to a specific extension. Type <prefix>help <cmd_name> to view information regarding a command.\n\n[ .Prefixes ] : {self.prefixes}\n\n"
                            msg += f"[ .Commands ]\n"
                            for command in ext.commands:
                                if command.ext == ext.ext:
                                    msg += f". {command.name}: {command.description}\n"

                            msg += "```"
                            return await ctx.reply(f"{msg}")
                    else:
                        for cmd in self.commands:
                            if name == cmd.name.lower():
                                msg = f"```ini\n[ {self.user.username} Selfbot ]\n"
                                msg += f"[ {self.user} ]\n\nType <prefix>help <ext_name> to view commands relating to a specific extension. Type <prefix>help <cmd_name> to view information regarding a command.\n\n[ .Prefixes ] : {self.prefixes}\n\n"
                                msg += f"[ .{cmd.name} ]\n"
                                msg += f"[ Description ] :  {cmd.description} \n"
                                msg += f"[ Long Description ] :\n{cmd.func.__doc__}\n"
                                msg += f"[ Aliases ] : {cmd.aliases} \n"
                                args = inspect.signature(cmd.func)
                                msg += f"\n[ Example Usage ] :\n[ {self.prefixes[0]}{cmd.aliases[0]}"
                                for arg in args.parameters.keys():
                                    if arg in ["self", "ctx"]:
                                        continue
                                    msg += f" <{arg}>"
                                msg += " ]"

                                msg += "```"
                                return await ctx.reply(f"{msg}")
                        for ext in self.extensions:
                            for cmd in ext.commands:
                                if name == cmd.name.lower():
                                    msg = f"```ini\n[ {self.user.username} Selfbot ]\n"
                                    msg += f"[ {self.user} ]\n\nType <prefix>help <ext_name> to view commands relating to a specific extension. Type <prefix>help <cmd_name> to view information regarding a command.\n\n[ .Prefixes ] : {self.prefixes}\n\n"
                                    msg += f"[ .{cmd.name} ]\n"
                                    msg += f"[ Description ] :  {cmd.description} \n"
                                    msg += (
                                        f"[ Long Description ] :\n{cmd.func.__doc__}\n"
                                    )
                                    msg += f"[ Aliases ] :  {cmd.aliases} \n"
                                    args = inspect.signature(cmd.func)
                                    msg += f"\n[ Example Usage ] :\n[ {self.prefixes[0]}{cmd.aliases[0]}"
                                    for arg in args.parameters.keys():
                                        if arg in ["self", "ctx"]:
                                            continue
                                        msg += f" <{arg}>"
                                    msg += " ]"

                                    msg += "```"
                                    return await ctx.reply(f"{msg}")

        if self.eval:
            # print("EVAL ACTIVATE???")
            def clean_code(content):
                if content.startswith("```") and content.endswith("```"):
                    return "\n".join(content.split("\n")[1:])[:-3]
                else:
                    return content



            @self.cmd(description="Executes and runs code", aliases=["exec"])
            async def eval(ctx, *, code):
                """Runs python code via exec, intended for experienced usage. This can be DANGEROUS if you do not know what you are doing, use with caution."""
                if code.startswith("```"):
                    code = clean_code(code)
                # print("EVAL ACTIVATE???")
                envs = {
                    "bot": self,
                    "ctx": ctx,
                    "selfcord": sys.modules[__name__],
                    "__import__": __import__
                }
                
                try:
                    with contextlib.redirect_stdout(io.StringIO()) as f:
                        await aexec(code, local=envs)
                        result = f"```{f.getvalue()}\n```"
                except Exception as e:
                    error = "".join(format_exception(e, e, e.__traceback__))
                    result = f"```\n{error}```"

                await ctx.reply(result)

    def on(self, event: str, mass_token: bool = False):
        """Decorator for events

        Args:
            event (str): The event to check for
        """

        def decorator(coro):
            if not inspect.iscoroutinefunction(coro):
    
                raise Exception("Not a coroutine")
            else:
                self._events[event].append(Event(name=event, coro=coro, ext=None, mass_token=mass_token))

                def wrapper(*args, **kwargs):
                    result = self._events[event].append(
                        Event(name=event, coro=coro, ext=None, mass_token=mass_token)
                    )

                    return result

                return wrapper

        return decorator

    async def emit(self, event, *args, **kwargs):
        """Used to essentially push values to the decorator when the event fires

        Args:
            event (str): The event name
        """
        on_event = f"on_{event}"

        # try:
        if hasattr(self, on_event):
            await getattr(self, on_event)(*args, **kwargs)
        
        if event in self._events.keys():
            
            for Event in self._events[event]:
                if len(Event.coro.__code__.co_varnames) == 0:
                    asyncio.create_task(Event.coro())
                elif Event.coro.__code__.co_varnames[0] == "self":
                    asyncio.create_task(Event.coro(Event.ext, *args, **kwargs))

                else:
                    asyncio.create_task(Event.coro(*args, **kwargs))

    def cmd(self, description="", aliases=[], mass_token: bool = False):
        """Decorator to add commands for the bot

        Args:
            description (str, optional): Description of command. Defaults to "".
            aliases (list, optional): Alternative names for command. Defaults to [].

        Raises:
            RuntimeWarning: If you suck and don't use a coroutine
        """
        if isinstance(aliases, str):
            aliases = [aliases]

        def decorator(coro):
            name = coro.__name__
            if not inspect.iscoroutinefunction(coro):
   
                raise Exception("Not a coroutine")
                return
            else:
                cmd = Command(
                    name=name, description=description, aliases=aliases, func=coro, mass_token=mass_token
                )
                self.commands.add(cmd)
            return cmd

        return decorator
    

    async def load_tokens(self, tokens: list, prefixes: list[str] = ["!"], eval: bool = False):
        rmv = []
        for token in tokens:

            mass_bot = self.__class__(prefixes=prefixes, token_leader=self.user.id, eval=eval, inbuilt_help=False, )

            @mass_bot.cmd()
            async def help(ctx):
                desc = "```\n"
                for cmd in mass_bot.commands:
                    desc += f"{cmd.name}  :  {cmd.description}\n"
                desc += "```"
                await ctx.reply(desc)

            for cmd in self.commands:
                if cmd.mass_token:
                    rmv.append(cmd)
                    mass_bot.commands.add(cmd)

            for ext in self.extensions:
                for name, value in ext._events.items():
                    for item in value:
                        if item.mass_token:
                            
                            mass_bot._events[name].append(item)

            for name, value in self._events.items():
                for item in value:
                    if item.mass_token:
                        
                        mass_bot._events[name].append(item)
            
            self.bots.append(mass_bot)

        for rm in rmv:
            try:
                self.commands.remove(rm)
            except:
                pass

        for token, bot in zip(tokens, self.bots):
            asyncio.create_task(bot.runner(token, True))

    def add_cmd(self, coro, description="", aliases=[]):
        """
        Function to add commands manually without decorator
        Args:
            coro (coroutine): The function to add
            description (str, optional): Description of command. Defaults to "".
            aliases (list, optional): Alternative names for command. Defaults to [].

        Raises:
            RuntimeWarning: If you suck and don't use a coroutine
        """
        if isinstance(aliases, str):
            aliases = [aliases]
        name = coro.__name__
        if not inspect.iscoroutinefunction(coro):

            raise Exception("Not a coroutine")
        else:
            cmd = Command(
                name=name, description=description, aliases=aliases, func=coro
            )
            self.commands.add(cmd)

    async def load_extension(self, name: Optional[str] = None, url: Optional[str] = None, dir: Optional[str] = None):
        """
        Load various extensions/plugins/cogs if any.

        Args:
            name (str): Name of the extension to load
            url (str): URL you want to load
            dir (str): Directory you want to load

        Raises:
            ModuleNotFoundError: If you suck and don't know the name of what you want to load
        """
        if name is None and url is None:
            return
        if url is not None:
            async with aiohttp.ClientSession() as session:
                async with session.get(url) as resp:
                    text = await resp.text()

            if dir is None:
                path = os.path.basename(urlparse(url).path)

                if path.endswith(".py"):
                    if os.path.exists(path):
          
                        return

                    lines = text.splitlines()
                    async with aiofiles.open(path, "a+") as f:
                        for line in lines:
                            await f.write(f"{line}\n")

                    name = f"{os.path.basename(urlparse(url).path)[:-3]}"

                else:
 
                    return

            else:
                path = f"{dir}/{os.path.basename(urlparse(url).path)}"
                
                if path.endswith(".py"):
                    if os.path.exists(path):
               
                        return
                    if not os.path.exists(dir):
                        os.makedirs(dir)
                    lines = text.splitlines()
                    async with aiofiles.open(path, "a+") as f:
                        for line in lines:
                            await f.write(f"{line}\n")

                    name = f"{dir}.{os.path.basename(urlparse(url).path)[:-3]}"

                else:
        
                    return

                
                
        try:
            name = importlib.util.resolve_name(name, None) # type: ignore
        except Exception as e:
            raise e

        spec = importlib.util.find_spec(name) # type: ignore

        lib = importlib.util.module_from_spec(spec) # type: ignore

        try:
            spec.loader.exec_module(lib)
        except Exception as e:

            return
        try:
            ext = getattr(lib, "Ext")
        except Exception as e:
            raise e

        # Creates an Extension - ext in this case refers to the Ext class used for initialisation
        ext = Extension(
            name=ext.name,
            description=ext.description,
            ext=ext(self),
            _events=ext._events,
        )
        self.extensions.add(ext)
        try:
            for name, event in ext._events.items():
                for ext_event in event:
                    self._events[name].append(
                        Event(name=name, coro=ext_event.coro, ext=ext.ext)
                    )
        except Exception as e:
            raise e

    async def logout(self):
        await self.gateway.close()

    async def change_presence(self, activity: list[dict], status: Literal["online", "idle", "dnd", "invisible"] = "online", afk: bool = False):
        d = {"op": 3, "d": {"activities": activity, "status": status, "afk": afk, "since": int(time.time())}}

        await self.gateway.send_json(d)


    async def process_commands(self, msg):
        """
        What is called in order to actually get command input and run commands

        Args:
            msg (str): The message containing command
        """
        context = Context(msg, self)

        asyncio.create_task(context.invoke())
    
    def fetch_message(self, message_id: str) -> Optional[Message]:
        if message_id is None:
            return
        return self.cached_messages.get(message_id)

    def fetch_user(self, user_id: str) -> Optional[User]:
        if user_id is None:
            return
        return self.cached_users.get(user_id)
    
    def get_voice(self) -> Optional[Voice]:
        return self.gateway.voice

    def get_channel(self, channel_id: str) -> Optional[Messageable|Callable]:
        if channel_id is None:
            return
        return self.cached_channels.get(channel_id)
    
    def get_dm(self, user_id: str) -> Optional[DMChannel]:
        for channel in self.cached_channels.values():
            if isinstance(channel, DMChannel):
               
                if channel.recipient.id == user_id:
                    return channel


    # Cry it's O(N) - max 100 guilds so it's cool
    def fetch_guild(self, guild_id: str) -> Optional[Guild]:
        if guild_id is None:
            return
        for guild in self.user.guilds:
            if guild.id == guild_id:
                return guild

    
    def fetch_role(self, role_id: str) -> Optional[Role]:
        if role_id is None:
            return
        for guild in self.user.guilds:
            for role in guild.roles:
                if role.id == role_id:
                    return role



    async def get_user(self, user_id: str) -> Optional[User]:
        """
        Function to retrieve user data. Probably need to be friends with them to retrieve the details.

        Args:
            user_id (Str): ID of the other user.

        Returns:

            User: The User object
        """
        data = await self.http.request(method="get", endpoint=f"/users/{user_id}")
        if data is not None:
            user = User(data, bot=self)
            self.cached_users[user.id] = user
            return user
        return
    
    async def join_invite(self, invite_code: str) -> Optional[Guild|GroupChannel]:
        json = await self.http.request(method="post", endpoint=f"/invites/{invite_code}", json={})

        if json is not None:
            if "guild" in json:
                guild = Guild(json["guild"], bot=self)
                self.user.guilds.append(guild)
                return guild
            elif "channel" in json:
                channel = GroupChannel(json["channel"], bot=self)
                self.user.private_channels.append(channel)
                self.cached_channels[channel.id] = channel
                return channel


 



from __future__ import annotations
from typing import Optional, TYPE_CHECKING, Literal
from .assets import Asset
from .permissions import Permission

if TYPE_CHECKING:
    from ..bot import Bot
    from .guild import Guild, Role
    from .channels import Convert, Messageable, DMChannel


# Realise this might be fucked because my subclassism didn't work with channels
# So basically guys my epic OOP magic didn't work this is indeed fucked
# Time to copy and paste everything! Actually methods should be fine, attrs for some reason don't wanna work

class Status:
    def __init__(self, payload: dict):
        self.update(payload)

    def __str__(self):
        return f"{self.platforms} // {self.status}"

    def update(self, payload: dict):
        self.platforms = [k for k in payload.keys()]
        self.status = [v for v in payload.values()]


class Profile():
    def __init__(self, id: str, payload: dict):
        self.id = id
        self.update(payload)

    def update(self, payload: dict):
        profile = payload["user_profile"]
        self.bio: Optional[str] = profile.get("bio")
        self.accent_color: Optional[str] = profile.get("accent_color")
        self.pronouns: Optional[str] = profile.get("pronouns")
        self.profile_effect: Optional[str] = profile.get("profile_effect")
        self.banner: Optional[Asset] = (
            Asset(self.id, payload["banner"]).from_avatar()
            if payload.get("banner") is not None and self.id is not None
            else None
        )
        self.theme_colors: Optional[list[int]] = payload.get("theme_colors")
        self.popout_animation_particle_type: Optional[str] = payload.get("popout_animation_particle_type")
        self.emoji: Optional[str] = payload.get("emoji")
        self.mutual_guilds = payload.get("mutual_guilds", [])
        


class User:
    def __init__(self, payload: dict, bot: Bot):
        self.bot = bot
        self.http = bot.http
        self.update(payload)


    def __str__(self):
        return f"{self.username}#{self.discriminator} ({self.id})"

    def __repr__(self):
        return f"<User id={self.id} name={self.display_name} discriminator={self.discriminator}>"


    def _remove_null(self, payload: dict):
        return {key: value for key, value in payload.items() if value is not None}
    

    
    async def profile(self) -> Optional[Profile]:
        resp = await self.http.request("GET", f"/users/{self.id}/profile?with_mutual_guilds=True")
        if resp is not None:
            return Profile(resp['user']['id'], resp)
        else:
            resp = await self.http.request("GET", f"/users/{self.id}/profile?with_mutual_guilds=False")
            if resp is not None:
                return Profile(resp['user']['id'], resp)


    def update(self, payload: dict):
        self.username: Optional[str] = payload.get("username")
        self.status: Optional[str] = payload.get("status")
        self.client_status: Optional[Status] = (
            Status(payload['client_status']) 
            if payload.get("client_status") is not None
            else None
        )
        self.broadcast = payload.get("broadcast")
        self.activities = payload.get("activities")
        self.id: Optional[str] = payload.get("id") or payload.get("user_id")

        self.discriminator: Optional[str] = payload.get("discriminator")
        self.global_name: Optional[str] = payload.get("global_name")
        self.avatar: Optional[Asset] = (
            Asset(self.id, payload["avatar"]).from_avatar()
            if payload.get("avatar") is not None and self.id is not None
            else None
        )
        self.banner: Optional[Asset] = (
            Asset(self.id, payload["banner"]).from_avatar()
            if payload.get("banner") is not None and self.id is not None
            else None
        )
        self.banner_color: Optional[str] = payload.get("banner_color")
        self.accent_color: Optional[str] = payload.get("accent_color")
        self.global_name: Optional[str] = payload.get("global_name")
        self.flags: int = payload.get("flags", 0)
        self.avatar_decoration: Optional[str] = payload.get("avatar_decoration")
        self.is_bot = payload.get("bot", False)
        self.premium_since = payload.get("premium_since")

        self.session_id = payload.get("session_id")
        self.self_video = payload.get("self_video")
        self.suppress = payload.get("suppress")
        self.self_mute = payload.get("self_mute")
        self.self_deaf = payload.get("self_deaf")
        self.request_to_speak_timestamp = payload.get("request_to_speak_timestamp")
        self.mute = payload.get("mute")
        self.deaf = payload.get("deaf")
        

    def partial_update(self, payload: dict):
        for key, value in payload.items():
            if hasattr(self, key):
                if key == "banner":
                    setattr(self, key, (
                        Asset(self.id, payload["banner"]).from_avatar()
                        if payload.get("banner") is not None and self.id is not None
                        else None
                    ))
                elif key == "avatar":
                    setattr(self, key, (
                        Asset(self.id, payload["banner"]).from_avatar()
                        if payload.get("banner") is not None and self.id is not None
                        else None
                    ))
                elif key == "client_status":
                    setattr(self, key, (
                        Status(payload['client_status']) 
                        if payload.get("client_status") is not None
                        else None
                    ))
                elif key == "bot":
                    setattr(self, "is_bot", value)

                else:
                    setattr(self, key, value)


    async def friend(self):
        json = await self.http.request(
            "put", "/users/@me/relationships/" + self.id if self.id is not None else "", json={}
        )
        return User(json, self.bot)
    
    async def block(self):
        await self.http.request(
            "put", "/users/@me/relationships/" + self.id if self.id is not None else "", json={"type": 2}
        )

    async def reset_relationship(self):
        await self.http.request(
            "delete", "/users/@me/relationships/" + self.id if self.id is not None else "", json={}
        )

    async def create_dm(self) -> Optional[DMChannel]:
        json = await self.http.request(
            "post", "/users/@me/channels", json={"recipients": [self.id if self.id is not None else ""]}
        )
        # print(json)

        return Convert(json, self.bot)


class Client(User):
    def __init__(self, payload: dict, bot: Bot):
        self.bot = bot
        self.http = bot.http
        self.guilds: list[Guild] = []
        self.friends: list[User] = []
        self.blocked: list[User] = []
        self.private_channels: list[Messageable] = []

        super().__init__(payload, bot) 
        super().update(payload)
        self.update(payload) 

    def update(self, payload: dict):
        self.verified = payload.get("verified")
        self.purchased_flags = payload.get("purchased_flags")
        self.pronouns = payload.get("pronouns")
        self.premium_type = payload.get("premium_type")
        self.phone = payload.get("phone")
        self.nsfw = payload.get("nsfw_allowed")
        self.mobile = payload.get("mobile")
        self.desktop = payload.get("desktop")
        self.mfa = payload.get("mfa_enabled")

    def partial_update(self, payload: dict):
        payload = self._remove_null(payload)
        super().partial_update(payload)
        for key, value in payload.items():
            if hasattr(self, key):
                setattr(self, key, value)

    

    async def add_friend(self, username: str, discriminator: Optional[str] = None):
        json = await self.http.request(
            "post", "/users/@me/relationships", json={"username": username, "discriminator": discriminator}
        )
        if json is not None:
            return User(json, self.bot)
        

    async def add_friend_id(self, id: str):
        json = await self.http.request(
            "PUT", "/users/@me/relationships/" + id, json={}

        )
        if json is not None:
            return User(json, self.bot)


    async def create_group(self, recipients: list[str], name: str):
        json = await self.http.request(
            "post", "/users/@me/channels", json={"recipients": recipients, "name": name}
        )
        if json is not None:
            return Convert(json, self.bot)


    async def edit_display_name(self, global_name: str):
        await self.http.request(
            "PATCH", "/users/@me",
            json={"global_nane": global_name}
        )

    async def edit_pfp(self, avatar_url: str, animated: bool = False):
        await self.http.request(
            "PATCH", "/users/@me",
            json={"avatar": (await self.http.encode_image(avatar_url, animated))}
        )

    async def edit_banner(self, banner_url: str, animated: bool = False):
        await self.http.request(
            "PATCH", "/users/@me/profile",
            json={"avatar": (await self.http.encode_image(banner_url, animated))}
        )

    async def edit_bio(self, bio: str):
        await self.http.request(
            "PATCH", "/users/@me/profile",
            json={"bio": bio}
        )
    

    async def redeem_nitro(self, code: str):
        json = await self.http.request(
            "post", f"/entitlements/gift-codes/{code}/redeem" + code, json={}
        )
        return json
    
    async def edit_hypesquad(self, house: int | Literal["bravery", "brilliance", "balance"]):
        match house:
            case "bravery":
                house = 1
            case "brilliance":
                house = 2
            case "balance":
                house = 3
            case _:
                raise ValueError("Invalid house")
            
        await self.http.request(
            "post", "/hypesquad/online", json={"house_id": house}
        )
                                               
                                               
    

class Member(User):
    def __init__(self, payload: dict, bot: Bot):
        self.bot = bot
        self.http = bot.http
        super().__init__(payload, bot)
        super().update(payload)
        self.update(payload)

    @property
    def guild(self):
        return self.bot.fetch_guild(self.guild_id)


    def update(self, payload: dict):
        self.roles: list[Role] = []
        self.permissions = None
        for role in payload.get("roles", []):
            role = self.bot.fetch_role(role)
            if role is not None:
                self.roles.append(role)
        
        self.guild_id: Optional[str] = payload.get("guild_id")
        self.joined_at: Optional[str] = payload.get("joined_at")
        self.premium_since: Optional[str] = payload.get("premium_since")
        self.deaf: bool = payload.get("deaf", False)
        self.mute: bool = payload.get("mute", False)
        self.pending: Optional[bool] = payload.get("pending")
        self.nick: Optional[str] = payload.get("nick")
        self.communication_disabled_until: Optional[str] = payload.get("communication_disabled_until")
        
        
    def partial_update(self, payload: dict):
        payload = self._remove_null(payload)
        super().partial_update(payload)
        for key, value in payload.items():
            if hasattr(self, key):
                if key == "roles":
                    self.roles = []
                    for role in value:
                        role = self.bot.fetch_role(role)
                        if role is not None:
                            self.roles.append(role)
                else:
                    setattr(self, key, value)


    async def kick(self, reason: str = ""):
        await self.http.request(
            "DELETE", f"/guilds/{self.guild_id}/members/{self.id}",
            headers={"X-Audit-Log-Reason": reason}
        )

    async def ban(self, reason: str = "", delete_message_seconds: int = 0):
        await self.http.request(
            "PUT", f"/guilds/{self.guild_id}/bans/{self.id}",
            json={"delete_message_seconds": delete_message_seconds},
            headers={"X-Audit-Log-Reason": reason}
        )

    async def unban(self):
        await self.http.request(
            "DELETE",
            f"/guilds/{self.guild_id}/bans/{self.id}",
        )
    



